"""Main module for extracting metafeatures from time-series."""
import typing as t
import collections
import shutil
import time

import texttable
import numpy as np

import tspymfe._internal as _internal
import tspymfe._period as _period
import tspymfe._detrend as _detrend

_TypeSeqExt = t.Sequence[
    t.Tuple[str, t.Callable, t.Tuple[str, ...], t.Tuple[str, ...]]
]
"""Type annotation for a sequence of TypeExtMtdTuple objects."""


class TSMFE:
    """Core class for time-series metafeature extraction.

    Attributes
    ----------
    ts : :obj:`Sequence`
        Independent attributes of the dataset.

    groups : :obj:`tuple` of :obj:`str`
        Tuple object containing fitted meta-feature groups loaded in the model
        at instantiation.

    features : :obj:`tuple` of :obj:`str`
        Contains loaded meta-feature extraction method names available for
        meta-feature extraction, from selected metafeatures groups and features
        listed at instantiation.

        summary : :obj:`tuple` of :obj:`str`
            Tuple object which contains summary functions names for features
            summarization.
    """

    groups_alias = [("default", _internal.DEFAULT_GROUP)]

    def __init__(
        self,
        groups: t.Union[str, t.Iterable[str]] = "default",
        features: t.Union[str, t.Iterable[str]] = "all",
        summary: t.Union[str, t.Iterable[str]] = ("mean", "sd"),
        measure_time: t.Optional[str] = None,
        wildcard: str = "all",
        score: str = "rmse",
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        suppress_warnings: bool = False,
        random_state: t.Optional[int] = None,
    ) -> None:
        """Provides easy access for metafeature extraction from datasets.

        It expected that user first calls ``fit`` method after instantiation
        and then ``extract`` for effectively extract the selected metafeatures.
        Check reference [1]_ for more information.

        Parameters
        ----------
        groups : :obj:`Iterable` of :obj:`str` or :obj:`str`
            A collection or a single metafeature group name representing the
            desired group of metafeatures for extraction. Use the method
            ``valid_groups`` to get a list of all available groups.

            Setting with ``all`` enables all available groups.

            The value provided by the argument ``wildcard`` can be used to
            select all metafeature groups rapidly.

        features : :obj:`Iterable` of :obj:`str` or :obj:`str`, optional
            A collection or a single metafeature name desired for extraction.
            Keep in mind that the extraction only gathers features also in the
            selected ``groups``. Check this class ``features`` attribute to get
            a list of available metafeatures from selected groups, or use the
            method ``valid_metafeatures`` to get a list of all available
            metafeatures filtered by group. Alternatively, you can use the
            method ``metafeature_description`` to get or print a table with
            all metafeatures with its respectives groups and descriptions.

            The value provided by the argument ``wildcard`` can be used to
            select all features from all selected groups rapidly.

        summary : :obj:`Iterable` of :obj:`str` or :obj:`str`, optional
            A collection or a single summary function to summarize a group of
            metafeature measures into a fixed-length group of value, typically
            a single value. The values must be one of the following:

                1. ``mean``: Average of the values.
                2. ``sd``: Standard deviation of the values.
                3. ``count``: Computes the cardinality of the measure. Suitable
                   for variable cardinality.
                4. ``histogram``: Describes the distribution of the measured
                   values. Suitable for high cardinality.
                5. ``iq_range``: Computes the interquartile range of the
                   measured values.
                6. ``kurtosis``: Describes the shape of the measures values
                   distribution.
                7. ``max``: Results in the maximum value of the measure.
                8. ``median``: Results in the central value of the measure.
                9. ``min``: Results in the minimum value of the measure.
                10. ``quantiles``: Results in the minimum, first quartile,
                    median, third quartile and maximum of the measured values.
                11. ``range``: Computes the range of the measured values.
                12. ``skewness``: Describes the shape of the measure values
                    distribution in terms of symmetry.

            You can concatenate `nan` with the desired summary function name
            to use an alternative version of the same summary which ignores
            `nan` values. For instance, `nanmean` is the `mean` summary
            function which ignores all `nan` values, while 'naniq_range`
            is the interquartile range calculated only with valid (non-`nan`)
            values.

            If more than one summary function is selected, then all multivalued
            extracted metafeatures are summarized with each summary function.

            The particular value provided by the argument ``wildcard`` can be
            used to select all summary functions rapidly.

            Use the method ``valid_summary`` to get a list of all available
            summary functions.

        measure_time : :obj:`str`, optional
            Options for measuring the time elapsed during metafeature
            extraction. If this argument value is :obj:`NoneType`, no time
            elapsed is measured. Otherwise, this argument must be a :obj:`str`
            valued as one of the options below:

                1. ``avg``: average time for each metafeature (total time
                   divided by the feature cardinality, i.e., number of features
                   extracted by a single feature-extraction related method),
                   without summarization time.
                2. ``avg_summ``: average time for each metafeature (total time
                   of extraction divided by feature cardinality) including
                   required time for summarization.
                3. ``total``: total time for each metafeature, without
                   summarization time.
                4. ``total_summ``: total time for each metafeature including
                   the required time for summarization.

            The ``cardinality`` of the feature is the number of values
            extracted by a single calculation method.

            For example, ``mean`` feature has cardinality equal to the number
            of numeric features in the dataset, where ``cor`` (from
            ``correlation``) has cardinality equals to (N - 1)/2, where N is
            the number of numeric features in the dataset.

            The cardinality is used to divide the total execution time of that
            method if an option starting with ``avg`` is selected.

            If a summary method has cardinality higher than one (more than one
            value returned after summarization and, thus, creating more than
            one entry in the result lists) like, for example, ``histogram``
            summary method, then the corresponding time of this summary will be
            inserted only in the first correspondent element of the time list.
            The remaining entries are all filled with 0 value, to keep
            consistency between the size of all lists returned and index
            correspondence between they.

        wildcard : :obj:`str`, optional
            Value used as ``select all`` for ``groups``, ``features`` and
            ``summary`` arguments.

        score : :obj:`str`, optional
            Score metric used to extract ``landmarking`` metafeatures.

        num_cv_folds : :obj:`int`, optional
            Number of folds to create a Stratified K-Fold cross
            validation to extract the ``landmarking`` metafeatures.

        lm_sample_frac : :obj:`float`, optional
            Sample proportion used to produce the ``landmarking`` metafeatures.
            This argument must be in 0.5 and 1.0 (both inclusive) interval.

        suppress_warnings : :obj:`bool`, optional
            If True, then ignore all warnings invoked at the instantiation
            time.

        random_state : :obj:`int`, optional
            Random seed used to control random events. Keeps the experiments
            reproducible.

        Notes
        -----
            .. [1] Rivolli et al. "Towards Reproducible Empirical
               Research in Meta-Learning,".
               Rivolli et al. URL: https://arxiv.org/abs/1808.10406

        Examples
        --------

        Load a dataset

        >>> from sklearn.datasets import load_iris
        >>> from pymfe.tsmfe import TSMFE

        >>> data = load_iris()
        >>> y = data.target
        >>> ts = data.data

        Extract all measures

        >>> mfe = TSMFE()
        >>> mfe.fit(ts, y)
        >>> ft = mfe.extract()
        >>> print(ft)

        Extract general, local-stat and information-theoretic measures

        >>> mfe = TSMFE(groups=["general", "local-stat", "info-theory"])
        >>> mfe.fit(ts, y)
        >>> ft = mfe.extract()
        >>> print(ft)

        """
        self.groups = _internal.process_generic_set(
            values=groups,
            group_name="groups",
            groups_alias=TSMFE.groups_alias,
            wildcard=wildcard,
        )  # type: t.Tuple[str, ...]

        (
            self.groups,
            self.inserted_group_dep,
        ) = _internal.solve_group_dependencies(groups=self.groups)

        proc_feat = _internal.process_features(
            features=features,
            groups=self.groups,
            suppress_warnings=suppress_warnings,
            wildcard=wildcard,
        )  # type: t.Tuple[t.Tuple[str, ...], _TypeSeqExt, t.Tuple[str, ...]]

        self.features, self._metadata_mtd_ft, self.groups = proc_feat
        del proc_feat

        self.summary, self._metadata_mtd_sm = _internal.process_summary(
            summary, wildcard=wildcard
        )  # type: t.Tuple[t.Tuple[str, ...], _TypeSeqExt]

        self.timeopt = _internal.process_generic_option(
            value=measure_time, group_name="timeopt", allow_none=True
        )  # type: t.Optional[str]

        self.ts = None  # type: t.Optional[np.ndarray]
        self.ts_trend = None  # type: t.Optional[np.ndarray]
        self.ts_season = None  # type: t.Optional[np.ndarray]
        self.ts_residuals = None  # type: t.Optional[np.ndarray]
        self.ts_detrended = None  # type: t.Optional[np.ndarray]
        self.ts_deseasonalized = None  # type: t.Optional[np.ndarray]
        self.ts_period = -1

        self._custom_args_ft = None  # type: t.Optional[t.Dict[str, t.Any]]
        """User-independent arguments for ft. methods (e.g. ``ts``)"""

        self._custom_args_sum = None  # type: t.Optional[t.Dict[str, t.Any]]
        """User-independent arguments for summary functions methods."""

        self._precomp_args_ft = None  # type: t.Optional[t.Dict[str, t.Any]]
        """Precomputed common feature-extraction method arguments."""

        self._postprocess_args_ft = {}  # type: t.Dict[str, t.Any]
        """User-independent arguments for post-processing methods."""

        if random_state is None or isinstance(random_state, int):
            self.random_state = random_state
            np.random.seed(random_state)

        else:
            raise ValueError(
                'Invalid "random_state" argument ({0}). '
                "Expecting None or an integer.".format(random_state)
            )

        if isinstance(num_cv_folds, int):
            self.num_cv_folds = num_cv_folds

        else:
            raise ValueError(
                'Invalid "num_cv_folds" argument ({0}). '
                "Expecting an integer.".format(random_state)
            )

        if isinstance(lm_sample_frac, int):
            lm_sample_frac = float(lm_sample_frac)

        if isinstance(lm_sample_frac, float) and 0.5 <= lm_sample_frac <= 1.0:
            self.lm_sample_frac = lm_sample_frac

        else:
            raise ValueError(
                'Invalid "lm_sample_frac" argument ({0}). '
                "Expecting an float [0.5, 1].".format(random_state)
            )

        self.score = _internal.check_score(score, self.groups)

        # """Total time elapsed for precomputations."""
        self.time_precomp = -1.0

        # """Total time elapsed for metafeature extraction."""
        self.time_extract = -1.0

        # """Total time elapsed in total (precomp + extract.)"""
        self.time_total = -1.0

    def _call_summary_methods(
        self,
        feature_values: t.Sequence[_internal.TypeNumeric],
        feature_name: str,
        verbose: int = 0,
        suppress_warnings: bool = False,
        **kwargs,
    ) -> t.Tuple[
        t.List[str], t.List[t.Union[float, t.Sequence]], t.List[float]
    ]:
        """Invoke summary functions loaded in the model on given feature
        values.

        Parameters
        ----------
        feature_values : :obj:`sequence` of numerics
            Sequence containing values from feature-extraction methods.

        feature_name : :obj:`str`
            Name of the feature method used for produce the ``feature_value.``

        verbose : :obj:`int`, optional
            Select the verbosity level of the summarization process.
            If == 1, then print just the ending message, without a line break.
            If >= 2, then messages about the summarization process may be
            printed. Note that there is no relation between this argument and
            warnings (see ``suppress_warnings`` argument below).

        suppress_warnings : :obj:`bool`, optional
            If True, ignore all warnings invoked before and after summary
            method calls. The summary callables may still invoke warnings by
            itself and the user need to ignore them, if possible, via kwargs.

        kwargs:
            User-defined arguments for the summary callables.

        Returns
        -------
        :obj:`tuple`(:obj:`list`, :obj:`list`, :obj:`list`)
            A tuple containing three lists.

            The first field is the identifiers of each summarized value in the
            form ``feature_name.summary_mtd_name`` (i.e., the feature
            extraction name concatenated by the summary method name, separated
            by a dot). If the summary function return more than one value
            (cardinality greater than 1), then each value name have an extra
            concatenated id starting from 0 to differ between values (i.e.
            ``feature_name.summary_mtd_name.id``).

            The second field is the summarized values. Both lists have a 1-1
            correspondence by the index of each element (i.e., the value at
            index ``i`` in the second list has its identifier at the same index
            in the first list and vice-versa).

            The third field is a list with measured time wasted by each summary
            function. If the cardinality of the summary function is greater
            than 1, then the correspondent measured time is kept only in the
            first correspondent field, and the extra fields are filled with 0
            to keep the consistency of the size between all lists.

                Example:
                    ([``attr_ent.mean``, ``attr_ent.sd``], [0.98346, 0.34436])
                    is the return value for the feature `attr_end` summarized
                    by both ``mean`` and ``sd`` (standard deviation), giving
                    the values ``0.98347`` and ``0.34436``, respectively.
        """
        metafeat_vals = []  # type: t.List[t.Union[int, float, t.Sequence]]
        metafeat_names = []  # type: t.List[str]
        metafeat_times = []  # type: t.List[float]

        for cur_metadata in self._metadata_mtd_sm:
            sm_mtd_name, sm_mtd_callable, sm_mtd_args, _ = cur_metadata

            if verbose >= 2:
                print(
                    " {} Summarizing '{}' feature with '{}' summary "
                    "function...".format(
                        _internal.VERBOSE_BLOCK_MID_SYMBOL,
                        feature_name,
                        sm_mtd_name,
                    ),
                    end=" ",
                )

            sm_mtd_args_pack = _internal.build_mtd_kwargs(
                mtd_name=sm_mtd_name,
                mtd_args=sm_mtd_args,
                mtd_mandatory=set(),
                user_custom_args=kwargs.get(sm_mtd_name),
                inner_custom_args=self._custom_args_sum,
                suppress_warnings=suppress_warnings,
            )

            summarized_val, time_sm = _internal.timeit(
                _internal.summarize,
                feature_values,
                sm_mtd_callable,
                sm_mtd_args_pack,
            )

            if not suppress_warnings:
                _internal.check_summary_warnings(
                    value=summarized_val,
                    name_feature=feature_name,
                    name_summary=sm_mtd_name,
                )

            if isinstance(summarized_val, np.ndarray):
                summarized_val = summarized_val.flatten().tolist()

            if isinstance(
                summarized_val, collections.Sequence
            ) and not isinstance(summarized_val, str):
                metafeat_vals += summarized_val
                metafeat_names += [
                    ".".join((feature_name, sm_mtd_name, str(i)))
                    for i in range(len(summarized_val))
                ]
                metafeat_times += [time_sm] + (
                    (len(summarized_val) - 1) * [0.0]
                )

            else:
                metafeat_vals.append(summarized_val)
                metafeat_names.append(".".join((feature_name, sm_mtd_name)))
                metafeat_times.append(time_sm)

            if verbose >= 2:
                print("Done.")

        if verbose >= 2:
            print(
                " {} Done summarizing '{}' feature.".format(
                    _internal.VERBOSE_BLOCK_END_SYMBOL, feature_name
                )
            )

        return metafeat_names, metafeat_vals, metafeat_times

    def _call_feature_methods(
        self,
        verbose: int = 0,
        # enable_parallel: bool = False,
        suppress_warnings: bool = False,
        **kwargs,
    ) -> t.Tuple[
        t.List[str], t.List[t.Union[int, float, t.Sequence]], t.List[float]
    ]:
        """Invoke feature methods loaded in the model and gather results.

        The returned values are already summarized if needed.

        For more information, check ``extract`` method documentation for
        in-depth information about arguments and return value.
        """
        metafeat_vals = []  # type: t.List[t.Union[int, float, t.Sequence]]
        metafeat_names = []  # type: t.List[str]
        metafeat_times = []  # type: t.List[float]

        skipped_count = 0
        for ind, cur_metadata in enumerate(self._metadata_mtd_ft, 1):
            (
                ft_mtd_name,
                ft_mtd_callable,
                ft_mtd_args,
                ft_mandatory,
            ) = cur_metadata

            ft_name_without_prefix = _internal.remove_prefix(
                value=ft_mtd_name, prefix=_internal.MTF_PREFIX
            )

            try:
                ft_mtd_args_pack = _internal.build_mtd_kwargs(
                    mtd_name=ft_name_without_prefix,
                    mtd_args=ft_mtd_args,
                    mtd_mandatory=ft_mandatory,
                    user_custom_args=kwargs.get(ft_name_without_prefix),
                    inner_custom_args=self._custom_args_ft,
                    precomp_args=self._precomp_args_ft,
                    suppress_warnings=suppress_warnings,
                )

            except RuntimeError:
                # Not all method's mandatory arguments were satisfied.
                # Skip the current method.
                if verbose >= 2:
                    print(
                        "\nSkipped '{}' ({} of {}).".format(
                            ft_mtd_name, ind, len(self._metadata_mtd_ft)
                        )
                    )

                skipped_count += 1
                continue

            if verbose >= 2:
                print(
                    "\nExtracting '{}' feature ({} of {})...".format(
                        ft_mtd_name, ind, len(self._metadata_mtd_ft)
                    )
                )

            features, time_ft = _internal.timeit(
                _internal.get_feat_value,
                ft_mtd_name,
                ft_mtd_args_pack,
                ft_mtd_callable,
                suppress_warnings,
            )

            ft_has_length = isinstance(
                features, (np.ndarray, collections.Sequence)
            )

            if ft_has_length and self._timeopt_type_is_avg():
                time_ft /= len(features)

            if self._metadata_mtd_sm and ft_has_length:
                sm_ret = self._call_summary_methods(
                    feature_values=features,
                    feature_name=ft_name_without_prefix,
                    verbose=verbose,
                    suppress_warnings=suppress_warnings,
                    **kwargs,
                )

                summarized_names, summarized_vals, times_sm = sm_ret

                metafeat_vals += summarized_vals
                metafeat_names += summarized_names
                metafeat_times += self._combine_time(time_ft, times_sm)

            else:
                metafeat_vals.append(features)
                metafeat_names.append(ft_name_without_prefix)
                metafeat_times.append(time_ft)

            if verbose > 0:
                _internal.print_verbose_progress(
                    cur_progress=100 * ind / len(self._metadata_mtd_ft),
                    cur_mtf_name=ft_mtd_name,
                    item_type="feature",
                    verbose=verbose,
                )

        if verbose == 1:
            _t_num_cols, _ = shutil.get_terminal_size()
            print(
                "\r{:<{fill}}".format(
                    "Process of metafeature extraction finished.",
                    fill=_t_num_cols,
                )
            )

        if verbose >= 2 and skipped_count > 0:
            print(
                "\nNote: skipped a total of {} metafeatures, "
                "out of {} ({:.2f}%).".format(
                    skipped_count,
                    len(self._metadata_mtd_ft),
                    100 * skipped_count / len(self._metadata_mtd_ft),
                )
            )

        return metafeat_names, metafeat_vals, metafeat_times

    def _timeopt_type_is_avg(self) -> bool:
        """Checks if user selected time option is an ``average`` type."""
        return isinstance(self.timeopt, str) and self.timeopt.startswith(
            _internal.TIMEOPT_AVG_PREFIX
        )

    def _timeopt_include_summary(self) -> bool:
        """Checks if user selected time option includes ``summary`` time."""
        return isinstance(self.timeopt, str) and self.timeopt.endswith(
            _internal.TIMEOPT_SUMMARY_SUFFIX
        )

    def _combine_time(
        self, time_ft: float, times_sm: t.List[float]
    ) -> t.List[float]:
        """Treat time from feature extraction and summarization based in
        ``timeopt``.

        Parameters
        ----------
        time_ft : :obj:`float`
            Time necessary to extract some feature.

        times_sm : :obj:`list` of :obj:`float`
            List of values to summarize the metafeature value with each summary
            function.

        Returns
        -------
        :obj:`list`
            If ``timeopt`` attribute considers ``summary`` time (i.e., selected
            option ends with ``summ``), then these returned list values are the
            combination of times gathered in feature extraction and
            summarization methods. Otherwise, the list values are the value of
            ``time_ft`` copied ``len(times_sm)`` times, to keep consistency
            with the correspondence between the values of all lists returned by
            ``extract`` method.
        """
        total_time = np.array([time_ft] * len(times_sm))

        if self._timeopt_include_summary():
            total_time += times_sm

        # As seen in ``_call_summary_methods`` method documentation, zero-
        # valued elements are created to fill the time list to keep its size
        # consistent with another feature extraction related lists. In this
        # case, here they're kept zero-valued.
        total_time[np.array(times_sm) == 0.0] = 0.0

        return total_time.tolist()

    def fit(
        self,
        ts: t.Sequence[int],
        ts_period: t.Optional[int] = None,
        rescale: t.Optional[str] = None,
        rescale_args: t.Optional[t.Dict[str, t.Any]] = None,
        precomp_groups: t.Optional[str] = "all",
        wildcard: str = "all",
        suppress_warnings: bool = False,
        verbose: int = 0,
        **kwargs,
    ) -> "TSMFE":
        """Fits dataset into an TSMFE model.

        Parameters
        ----------
        ts : :obj:`Sequence`
            Predictive attributes of the dataset.

        ts_period : int, optional
            Period of the time-series. If not given, it will be estimated
            from the autocorrelation function.

        rescale : :obj:`str`, optional
            If :obj:`NoneType`, the model keeps all numeric data with its
            original values. Otherwise, this argument can assume one of the
            string options below to rescale all numeric values:

                1. ``standard``: set numeric data to zero mean, unit variance.
                   Also known as ``z-score`` normalization. Check the
                   documentation of ``sklearn.preprocessing.StandardScaler``
                   for in-depth information.

                2. `'min-max``: set numeric data to interval [a, b], a < b. It
                   is possible to define values to ``a`` and ``b`` using
                   argument ``rescale_args``. The default values are a = 0.0
                   and b = 1.0. Check ``sklearn.preprocessing.MinMaxScaler``
                   documentation for more information.

                3. ``robust``: rescale data using statistics robust to the
                   presence of outliers. For in-depth information, check
                   documentation of ``sklearn.preprocessing.RobustScaler``.

        rescale_args : :obj:`dict`, optional
            Dictionary containing parameters for rescaling data. Used only if
            ``rescale`` argument is not :obj:`NoneType`. These dictionary keys
            are the parameter names as strings and the values, the
            corresponding parameter value.

        precomp_groups : :obj:`str`, optional
            Defines which metafeature groups common values should be cached to
            share among various meta-feature extraction related methods (e.g.
            ``classes``, or ``covariance``). This argument may speed up
            meta-feature extraction but also consumes more memory, so it may
            not be suitable for huge datasets.

        wildcard : :obj:`str`, optional
            Value used as ``select all`` for ``precomp_groups``.

        suppress_warnings : :obj:`bool`, optional
            If True, ignore all warnings invoked while fitting dataset.

        verbose : :obj:`int`, optional
            Defines the level of verbosity for the fit method. If `1`, then
            print a progress bar related to the precomputations. If `2` or
            higher, then log every step of the fitted data transformations and
            the precomputation steps.

        **kwargs:
            Extra custom arguments to the precomputation methods. Keep in
            mind that those values may even replace internal custom parameters,
            if the name matches. Use this resource carefully.

            Hint: you can check which are the internal custom arguments by
            verifying the values in '._custom_args_ft' attribute after the
            model is fitted.

            This argument format is {'parameter_name': parameter_value}.

        Returns
        -------
        self

        Raises
        ------
        ValueError
            If the number of rows of ts and y length does not match.
        TypeError
            If ts is neither a :obj:`list` or a :obj:`np.ndarray` object.

        """
        rescale = _internal.process_generic_option(
            value=rescale, group_name="rescale", allow_none=True
        )

        if verbose >= 2:
            print("Fitting data into model... ", end="")

        self.ts = _internal.check_data(ts)

        if rescale is not None:
            self.ts = _internal.rescale_data(
                data=self.ts.reshape(-1, 1), option=rescale, args=rescale_args
            ).ravel()

        if verbose >= 2:
            print("Done.")
            print("Getting time-series period... ", end="")

        self.ts_period = _period.get_ts_period(ts=self.ts, ts_period=ts_period)

        if verbose >= 2:
            print("Done (got period {}).".format(ts_period))
            print("Started time-series decomposition... ", end="")

        _ts_components = _detrend.decompose(self.ts, ts_period=self.ts_period)

        self.ts_trend, self.ts_season, self.ts_residuals = _ts_components

        self.ts_detrended = self.ts - self.ts_trend
        self.ts_deseasonalized = self.ts - self.ts_season

        if verbose >= 2:
            print("Done.")

        # Custom arguments for metafeature extraction methods
        self._custom_args_ft = {
            "ts": self.ts,
            "ts_trend": self.ts_trend,
            "ts_season": self.ts_season,
            "ts_residuals": self.ts_residuals,
            "ts_detrended": self.ts_detrended,
            "ts_deseasonalized": self.ts_deseasonalized,
            "ts_period": self.ts_period,
            "num_cv_folds": self.num_cv_folds,
            "lm_sample_frac": self.lm_sample_frac,
            "score": self.score,
            "random_state": self.random_state,
        }

        if verbose >= 2:
            print("Started precomputation process.")

        _time_start = time.time()

        # Custom arguments from preprocessing methods
        self._precomp_args_ft = _internal.process_precomp_groups(
            precomp_groups=precomp_groups,
            groups=self.groups,
            wildcard=wildcard,
            suppress_warnings=suppress_warnings,
            verbose=verbose,
            **{**self._custom_args_ft, **kwargs},
        )

        self.time_precomp = time.time() - _time_start

        if verbose >= 2:
            print(
                "\nFinished precomputation process.",
                " {} Total time elapsed: {:.8f} seconds".format(
                    _internal.VERBOSE_BLOCK_MID_SYMBOL, self.time_precomp
                ),
                " {} Got a total of {} precomputed values.".format(
                    _internal.VERBOSE_BLOCK_END_SYMBOL,
                    len(self._precomp_args_ft),
                ),
                sep="\n",
            )

        # Custom arguments for postprocessing methods
        self._postprocess_args_ft = {
            "inserted_group_dep": self.inserted_group_dep,
        }

        # Custom arguments for summarization methods
        self._custom_args_sum = {
            "ddof": 1,
        }

        return self

    def extract(
        self,
        verbose: int = 0,
        enable_parallel: bool = False,
        suppress_warnings: bool = False,
        **kwargs,
    ) -> t.Tuple[t.Sequence, ...]:
        """Extracts metafeatures from the previously fitted dataset.

        Parameters
        ----------
        verbose : :obj:`int`, optional
            Defines the verbosity level related to the metafeature extraction.
            If == 1, show just the current progress, without line breaks.
            If >= 2, print all messages related to the metafeature extraction
            process.

            Note that warning messages are not affected by this option (see
            ``suppress_warnings`` argument below).

        enable_parallel : :obj:`bool`, optional
            If True, then the meta-feature extraction is done with
            multi-processes. Currently, this argument has no effect by now
            (to be implemented).

        suppress_warnings : :obj:`bool`, optional
            If True, do not show warnings about unknown user custom parameters
            for feature extraction and summary methods passed via kwargs. Note
            that both feature extraction and summary methods may still raise
            warnings by itself.

        kwargs:
            Used to pass custom arguments for both feature-extraction and
            summary methods. The expected format is the following:

            {``mtd_name``: {``arg_name``: arg_value, ...}, ...}

            In words, the key values of ``**kwargs`` should be the target
            methods which receives the custom arguments, and each method has
            another dictionary containing customs method argument names as keys
            and their correspondent values, as values. See ``Examples``
            subsection for a clearer explanation.

            For more information see Examples.

        Returns
        -------
        :obj:`tuple`(:obj:`list`, :obj:`list`)
            A tuple containing two lists (if ``measure_time`` is None.)

            The first field is the identifiers of each summarized value in the
            form ``feature_name.summary_mtd_name`` (i.e., the feature
            extraction name concatenated by the summary method name, separated
            by a dot).

            The second field is the summarized values.

            Both lists have a 1-1 correspondence by the index of each element
            (i.e., the value at index ``i`` in the second list has its
            identifier at the same index in the first list and vice-versa).

            Example:
                ([``attr_ent.mean``, ``attr_ent.sd``], [``0.983``, ``0.344``])
                is the return value for the feature ``attr_end`` summarized by
                both ``mean`` and ``sd`` (standard deviation), giving the valu-
                es ``0.983`` and ``0.344``, respectively.

            if ``measure_time`` is given during the model instantiation, a
            third list will be returned with the time spent during the
            calculations for the corresponding (by index) metafeature.

        Raises
        ------
        TypeError
            If calling ``extract`` method before ``fit`` method.

        Examples
        --------
        Using kwargs. Option 1 to pass ft. extraction custom arguments:

        >>> args = {
        >>> 'sd': {'ddof': 2},
        >>> '1NN': {'metric': 'minkowski', 'p': 2},
        >>> 'leaves': {'max_depth': 4},
        >>> }

        >>> model = TSMFE().fit(ts=data)
        >>> result = model.extract(**args)

        Option 2 (note: metafeatures with name starting with numbers are not
        allowed!):

        >>> model = TSMFE().fit(ts=data)
        >>> res = extract(sd={'ddof': 2}, leaves={'max_depth': 4})

        """
        if self.ts is None:
            raise TypeError(
                "Fitted data not found. Call " '"fit" method before "extract".'
            )

        if not isinstance(self.ts, np.ndarray):
            self.ts = _internal.check_data(self.ts)

        if verbose >= 2:
            print("Started the metafeature extraction process.")

        _time_start = time.time()

        results = self._call_feature_methods(
            verbose=verbose,
            enable_parallel=enable_parallel,
            suppress_warnings=suppress_warnings,
            **kwargs,
        )  # type: t.Tuple[t.List, ...]

        _internal.post_processing(
            results=results,
            groups=self.groups,
            suppress_warnings=suppress_warnings,
            **self._postprocess_args_ft,
            **kwargs,
        )

        self.time_extract = time.time() - _time_start
        self.time_total = self.time_extract + self.time_precomp

        if results and results[0]:
            # Sort results by metafeature name
            results = tuple(
                map(
                    list, zip(*sorted(zip(*results), key=lambda item: item[0]))
                )
            )

        res_names, res_vals, res_times = results

        if verbose >= 2:
            _ext_t_pct = 100 * self.time_extract / self.time_total
            print(
                "\nMetafeature extraction process done.",
                " {} Time elapsed in total (precomputations + extraction): "
                "{:.8f} seconds.".format(
                    _internal.VERBOSE_BLOCK_MID_SYMBOL, self.time_total
                ),
                " {} Time elapsed for extractions: {:.8f} seconds ({:.2f}% "
                "from the total).".format(
                    _internal.VERBOSE_BLOCK_MID_SYMBOL,
                    self.time_extract,
                    _ext_t_pct,
                ),
                " {} Time elapsed for precomputations: {:.8f} seconds "
                "({:.2f}% from the total).".format(
                    _internal.VERBOSE_BLOCK_MID_SYMBOL,
                    self.time_precomp,
                    100 - _ext_t_pct,
                ),
                " {} Total of {} values obtained.".format(
                    _internal.VERBOSE_BLOCK_END_SYMBOL, len(res_vals)
                ),
                sep="\n",
            )

        if self.timeopt:
            return res_names, res_vals, res_times

        return res_names, res_vals

    def _extract_with_bootstrap(
        self,
        extractor: "TSMFE",
        sample_num: int,
        arguments_fit: t.Dict[str, t.Any],
        arguments_extract: t.Dict[str, t.Any],
        verbose: int = 0,
    ) -> t.Tuple[np.ndarray, ...]:
        """Extract metafeatures using bootstrapping."""
        if self.ts is None:
            raise TypeError(
                "Fitted data not found. Please call 'fit' " "method first."
            )

        def _handle_extract_ret(
            res: t.Tuple[np.ndarray, ...],
            args: t.Tuple[t.Sequence, ...],
            it_num: int,
        ) -> t.Tuple[np.ndarray, ...]:
            """Handle each .extraction method return value."""
            mtf_names, mtf_vals, mtf_time = res

            if not self.timeopt:
                cur_mtf_names, cur_mtf_vals = args

            else:
                cur_mtf_names, cur_mtf_vals, cur_mtf_time = args

            if mtf_names.size:
                mtf_vals[:, it_num] = cur_mtf_vals

                if self.timeopt:
                    mtf_time += cur_mtf_time

            else:
                mtf_names = np.asarray(cur_mtf_names, dtype=str)
                mtf_vals = np.zeros(
                    (len(cur_mtf_vals), sample_num), dtype=float
                )
                mtf_vals[:, 0] = cur_mtf_vals

                if self.timeopt:
                    mtf_time = np.asarray(cur_mtf_time, dtype=float)

            return mtf_names, mtf_vals, mtf_time

        res = 3 * (np.array([]),)

        if self.random_state is None:
            # Enforce pseudo-random behaviour to avoid previously set
            # random seeds out of this context
            np.random.seed()

        bootstrap_random_state = (
            self.random_state
            if self.random_state is not None
            else np.random.randint(2 ** 20 - 1)
        )

        for it_num in np.arange(sample_num):
            if verbose > 0:
                print(
                    "Extracting from sample dataset {} of {} ({:.2f}%)...".format(
                        1 + it_num,
                        sample_num,
                        100.0 * (1 + it_num) / sample_num,
                    )
                )

            # Note: setting random state to prevent same sample indices due
            # to random states set during fit/extraction
            np.random.seed(bootstrap_random_state)
            bootstrap_random_state += 1

            sample_inds = np.random.randint(
                self.ts.shape[0], size=self.ts.shape[0]
            )

            ts_sample = self.ts[sample_inds, :]

            extractor.fit(ts_sample, **arguments_fit)

            res = _handle_extract_ret(
                res=res,
                args=extractor.extract(**arguments_extract),
                it_num=it_num,
            )

            if verbose > 0:
                print(
                    "Done extracting from sample dataset {}.\n".format(
                        1 + it_num
                    )
                )

        return res

    def extract_with_confidence(
        self,
        sample_num: int = 128,
        confidence: t.Union[float, t.Sequence[float]] = 0.95,
        return_avg_val: bool = True,
        arguments_fit: t.Optional[t.Dict[str, t.Any]] = None,
        arguments_extract: t.Optional[t.Dict[str, t.Any]] = None,
        verbose: int = 0,
    ) -> t.Tuple[t.List, ...]:
        """Extract metafeatures with confidence intervals.

        To build the confidence intervals, each metafeature is extracted
        ``sample_num`` times from a distinct dataset built from the
        fitted data using bootstrap.

        All configuration used by this method are from the configuration
        while instantiating the current model.

        Parameters
        ----------
        sample_num : int, optional
            Number of samples from the fitted data using bootstrap. Each
            metafeature will be extracted ``sample_num`` times.

        confidence : float or sequence of floats, optional
            Confidence level of the interval. Must be in (0.0, 1.0) range.
            If a sequence of confidence levels is given, a confidence
            interval will be extracted for all values. Each confidence
            interval will be calculated as [confidence/2, 1 - confidence/2].

        return_avg_vals : bool, optional
            If True, return the average value for both the metafeature
            values and the time elapsed for its extraction (if any
            ``measure_time`` option was chosen.) If False, then all
            extracted metafeature values are returned as a 2D numpy array
            of shape (`metafeature_num`, `sample_num`) (i.e., each row
            represents a distinct metafeature, and each column is the
            value of the corresponding metafeature extracted from a
            distinct sample dataset) and the time elapsed will be the
            sum of all extractions for each metafeature.

        arguments_fit : dict, optional
            Extra arguments for the fit method for each sampled dataset.
            See ``.fit`` method documentation for more information.

        arguments_extract : dict, optional
            Extra arguments for each metafeature extraction procedure.
            See ``.extract`` method documentation for more information.

        verbose : int, optional
            Verbosity level for this method. Please note that the
            verbosity level for both ``.fit`` and ``.extract`` methods
            performed within this method must be controlled separately
            using, respectively, ``arguments_fit`` and ``arguments_extract``
            parameters.

        Returns
        -------
        tuple of :obj:`np.ndarray`
            The same return value format of the ``extract`` method, appended
            with the confidence intervals as a new sequence of values in the
            form (interval_low_1, interval_low_2, ..., interval_high_(n-1),
            interval_high_n) for each corresponding metafeature, and with shape
            (`metafeature_num`, 2 * C), where `C` is the number of confidence
            levels given in ``confidence`` (i.e., the rows represents each
            metafeature and the columns each interval limit). This means that
            all interval lower limits are given first, and all the interval
            upper limits are grouped together afterwards. The sequence order
            of the interval limits follows the same sequence order of the
            confidence levels given in ``confidence``. For instance, if
            `confidence=[0.80, 0.90, 0.99]`, then the confidence intervals
            will be returned in the following order (for all metafeatures):
            (lower_0.80, lower_0.90, lower_0.99, upper_0.80, upper_0.90,
            upper_0.99).

            if ``return_avg_val`` is True, the metafeature values and the
            time elapsed for extraction for each item (if any ``measure_time``
            options was chosen) will be the average value between all
            extractions. Otherwise, all extracted metafeature values will be
            returned as a 2D numpy array (where each columns is from a distinct
            sampled dataset, and each row is a distinct metafeature), and the
            time elapsed will be the sum of all extractions for the
            corresponding metafeature.

        Raises
        ------
        ValueError
            If ``confidence`` is not in (0.0, 1.0) range.

        Notes
        -----
        The model used to fit and extract metafeatures for each sampled
        dataset is instantiated within this method and, therefore, this
        method does not affect the current model (if any) by any means.
        """
        _confidence = np.asarray(confidence, dtype=float)

        if np.any(np.logical_or(_confidence <= 0.0, _confidence >= 1.0)):
            raise ValueError(
                "'_confidence' must be in (0.0, 1.0) range (got "
                "{}.)".format(_confidence)
            )

        if self.random_state is not None:
            np.random.seed(self.random_state)

        if arguments_fit is None:
            arguments_fit = {}

        if arguments_extract is None:
            arguments_extract = {}

        # Note: the metafeature extraction random seed will be fixed due
        # to the random indices while bootstrapping the fitted data.
        _random_state = self.random_state if self.random_state else 1234

        if verbose > 0:
            print("Started metafeature extract with _confidence interval.")
            print("Random seed:")
            print(
                " {} For extractor model: {}{}".format(
                    _internal.VERBOSE_BLOCK_END_SYMBOL,
                    _random_state,
                    "" if self.random_state else " (chosen by default)",
                )
            )

            print(
                " {} For bootstrapping: {}".format(
                    _internal.VERBOSE_BLOCK_END_SYMBOL, self.random_state
                )
            )

        extractor = TSMFE(
            features=self.features,
            groups=self.groups,
            summary=self.summary,
            measure_time=self.timeopt,
            random_state=_random_state,
        )

        mtf_names, mtf_vals, mtf_time = self._extract_with_bootstrap(
            extractor=extractor,
            sample_num=sample_num,
            verbose=verbose,
            arguments_fit=arguments_fit,
            arguments_extract=arguments_extract,
        )

        if verbose > 0:
            print("Finished metafeature extract with _confidence interval.")
            print("Now getting _confidence intervals...", end=" ")

        _half_sig_level = 0.5 * (1.0 - _confidence)
        quantiles = np.hstack((_half_sig_level, 1.0 - _half_sig_level))
        mtf_conf_int = np.quantile(a=mtf_vals, q=quantiles, axis=1).T

        if verbose > 0:
            print("Done.")

        if return_avg_val:
            mtf_vals = np.nanmean(mtf_vals, axis=1)

        if self.timeopt:
            if return_avg_val:
                mtf_time /= sample_num

            return mtf_names, mtf_vals, mtf_time, mtf_conf_int

        return mtf_names, mtf_vals, mtf_conf_int

    @classmethod
    def valid_groups(cls) -> t.Tuple[str, ...]:
        """Return a tuple of valid metafeature groups.

        Notes
        -----
        The returned ``groups`` are not related to the groups fitted in
        the model in the model instantation. The returned groups are all
        available metafeature groups in the ``Pymfe`` package. Check the
        ``TSMFE`` documentation for deeper information.
        """
        return _internal.VALID_GROUPS

    @classmethod
    def valid_summary(cls) -> t.Tuple[str, ...]:
        """Return a tuple of valid summary functions.

        Notes
        -----
        The returned ``summaries`` are not related to the summaries fitted
        in the model in the model instantation. The returned summaries are
        all available in the ``Pymfe`` package. Check the documentation of
        ``TSMFE`` for deeper information.
        """
        return _internal.VALID_SUMMARY

    @classmethod
    def _check_groups_type(
        cls, groups: t.Optional[t.Union[str, t.Iterable[str]]]
    ) -> t.Set[str]:
        """Cast ``groups`` to a tuple of valid metafeature group names."""
        if groups is None:
            return set(_internal.VALID_GROUPS)

        groups = _internal.convert_alias(TSMFE.groups_alias, groups)

        return set(groups)

    @classmethod
    def _filter_groups(cls, groups: t.Set[str]) -> t.Set[str]:
        """Filter given groups by the available metafeature group names."""
        filtered_group_set = {
            group for group in groups if group in _internal.VALID_GROUPS
        }
        return filtered_group_set

    @classmethod
    def valid_metafeatures(
        cls,
        groups: t.Optional[t.Union[str, t.Iterable[str]]] = None,
    ) -> t.Tuple[str, ...]:
        """Return a tuple with all metafeatures related to given ``groups``.

        Parameters
        ----------
        groups : :obj:`Sequence` of :obj:`str` or :obj:`str`, optional:
            Can be a string such value is a name of a specific metafeature
            group (see ``valid_groups`` method for more information) or a
            sequence of metafeature group names. It can be also None, which
            in that case all available metafeature names will be returned.

        Returns
        -------
        :obj:`tuple` of :obj:`str`
            Tuple with all available metafeature names of the given ``groups``.

        Notes
        -----
        The returned ``metafeatures`` are not related to the groups or to the
        metafeatures fitted in the model in the model instantation. All the
        returned metafeatures are available in the ``Pymfe`` package. Check
        the ``TSMFE`` documentation for deeper information.
        """
        groups = TSMFE._check_groups_type(groups)
        groups = TSMFE._filter_groups(groups)

        deps = _internal.check_group_dependencies(groups)

        mtf_names = []  # type: t.List
        for group in groups.union(deps):
            class_ind = _internal.VALID_GROUPS.index(group)

            mtf_names += _internal.get_prefixed_mtds_from_class(
                class_obj=_internal.VALID_MFECLASSES[class_ind],
                prefix=_internal.MTF_PREFIX,
                only_name=True,
                prefix_removal=True,
            )

        return tuple(mtf_names)

    @classmethod
    def parse_by_group(
        cls,
        groups: t.Union[t.Sequence[str], str],
        extracted_results: t.Tuple[t.Sequence, ...],
    ) -> t.Tuple[t.List, ...]:
        """Parse the result of ``extract`` for given metafeature ``groups``.

        Can be used to easily separate the results of each metafeature
        group.

        Parameters
        ----------
        groups : :obj:`Sequence` of :obj:`str` or :obj:`str`
            Metafeature group names which the results should be parsed
            relative to. Use ``valid_groups`` method to check the available
            metafeature groups.

        extracted_results : :obj:`tuple` of :obj:`t.Sequence`
            Output of ``extract`` method. Should contain all outputed lists
            (metafeature names, values and elapsed time for extraction, if
            present.)

        Returns
        -------
        :obj:`tuple` of :obj:`str`
            Slices of lists of ``extracted_results``, selected based on
            given ``groups``.

        Notes
        -----
        The given ``groups`` are not related to the groups fitted in the
        model in the model instantation. Check ``valid_groups`` method to
        get a list of all available groups from the ``Pymfe`` package.
        Check the ``TSMFE`` documentation for deeper information about all
        these groups.
        """
        selected_indexes = _internal.select_results_by_classes(
            mtf_names=extracted_results[0],
            class_names=groups,
            include_dependencies=True,
        )

        filtered_res = (
            [seq[ind] for ind in selected_indexes] for seq in extracted_results
        )

        return tuple(filtered_res)

    @staticmethod
    def _parse_description(
        docstring: str, include_references: bool = False
    ) -> t.Tuple[str, str]:
        """Parse the docstring to get initial description and reference.

        Parameters
        ----------
        docstring : str
            An numpy docstring as ``str``.

        include_references : bool
            If True include a column with article reference.

        Returns
        -------
        tuple of str
            The initial docstring description in the first position and the
            reference in the second.

        """
        initial_description = ""  # type: str
        reference_description = ""  # type: str

        # get initial description
        split = docstring.split("\n\n")
        if split:
            initial_description = " ".join(split[0].split())

        # get reference description
        if include_references:
            aux = docstring.split("References\n        ----------\n")
            if len(aux) >= 2:
                split = aux[1].split(".. [")
                if len(split) >= 2:
                    del split[0]
                    for spl in split:
                        reference_description += (
                            "[" + " ".join(spl.split()) + "\n"
                        )

        return (initial_description, reference_description)

    @classmethod
    def metafeature_description(
        cls,
        groups: t.Optional[t.Union[str, t.Iterable[str]]] = None,
        sort_by_group: bool = False,
        sort_by_mtf: bool = False,
        print_table: bool = True,
        include_references: bool = False,
    ) -> t.Optional[t.Tuple[t.List[t.List[str]], str]]:
        """Print a table with groups, metafeatures and description.

        Parameters
        ----------
        groups : sequence of str or str, optional:
            Can be a string such value is a name of a specific metafeature
            group (see ``valid_groups`` method for more information) or a
            sequence of metafeature group names. It can be also None, which
            in that case all available metafeature names will be returned.

        sort_by_group: bool
            Sort table by meta-feature group name.

        sort_by_mtf: bool
            Sort table by meta-feature name.

        print_table : bool
            If True a table will be printed with the description, otherwise the
            table will be send by return.

        print_table : bool
            If True sort the table by metafeature name.

        include_references : bool
            If True include a column with article reference.

        Returns
        -------
        list of list
            A table with the metafeature descriptions or None.

        Notes
        -----
        The returned ``metafeatures`` are not related to the groups or to the
        metafeatures fitted in the model instantation. All the
        returned metafeatures are available in the ``Pymfe`` package. Check
        the ``TSMFE`` documentation for deeper information.
        """

        groups = TSMFE._check_groups_type(groups)
        groups = TSMFE._filter_groups(groups)

        deps = _internal.check_group_dependencies(groups)

        if not isinstance(sort_by_group, bool):
            raise TypeError("The parameter sort_by_group should be bool.")

        if not isinstance(sort_by_mtf, bool):
            raise TypeError("The parameter sort_by_mtf should be bool.")

        if not isinstance(print_table, bool):
            raise TypeError("The parameter print_table should be bool.")

        mtf_desc = [["Group", "Meta-feature name", "Description"]]
        if include_references:
            mtf_desc[0].append("Reference")

        for group in groups.union(deps):
            class_ind = _internal.VALID_GROUPS.index(group)

            mtf_names = _internal.get_prefixed_mtds_from_class(  # type: ignore  # type: ignore
                class_obj=_internal.VALID_MFECLASSES[class_ind],
                prefix=_internal.MTF_PREFIX,
                only_name=False,
                prefix_removal=True,
            )

            for name, method in mtf_names:  # type: ignore
                ini_desc, ref_desc = TSMFE._parse_description(
                    str(method.__doc__), include_references
                )
                mtf_desc_line = [group, name, ini_desc]
                mtf_desc.append(mtf_desc_line)

                if include_references:
                    mtf_desc_line.append(ref_desc)

        if sort_by_mtf:
            mtf_desc.sort(key=lambda i: i[1])

        if sort_by_group:
            mtf_desc.sort(key=lambda i: i[0])

        draw = texttable.Texttable().add_rows(mtf_desc).draw()
        if print_table:
            print(draw)
            return None
        return mtf_desc, draw



"""Module dedicated to global statistics time-series meta-features."""
import typing as t
import warnings

import numpy as np
import nolds
import scipy.stats

import tspymfe._period as _period
import tspymfe._utils as _utils
import tspymfe._summary as _summary


class MFETSGlobalStats:
    """Extract time-series meta-features from Global Statistics group."""

    @classmethod
    def precompute_period(cls, ts: np.ndarray, **kwargs) -> t.Dict[str, int]:
        """Precompute the time-series period.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``ts_period`` (:obj:`int`): time-series period.
        """
        precomp_vals = {}  # type: t.Dict[str, int]

        if "ts_period" not in kwargs:
            precomp_vals["ts_period"] = _period.get_ts_period(ts=ts)

        return precomp_vals

    @classmethod
    def ft_ioe_tdelta_mean(
        cls,
        ts: np.ndarray,
        step_size: float = 0.05,
        normalize: bool = True,
        differentiate: bool = False,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Mean change of interval length with iterative outlier exclusion.

        This method calculates, at each iteration, the mean of the differences
        of the timestamps of instances using the iterative outlier exclusion
        strategy.

        In the iterative outlier exclusion, a uniformly spaced set of
        thresholds over the time-series range is build and, for each iteration,
        it is calculated a statistic of the diference of the timestamp values
        of instances larger or equal than the current threshold.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        step_size : float, optional (default=0.05)
            Increase of the outlier threshold in each iteration. Must be a
            number strictly positive.

        normalize : bool, optional (default=True)
            If True, normalize the statistic in the [-1, 1] interval. If
            False, return the raw mean timestamp values.

        differentiate : bool, optional (default=False)
            If True, differentiate the timestamps before calculating each
            statistic. If False, all statistics will be calculated on the
            raw timestamps.

        Returns
        -------
        :obj:`np.ndarray`
            If `differentiate` is False, the mean value of outlier timestamps
            of all iterations of the iterative outlier exclusion process. If
            `differentiate` is True, the mean value of the timestamps interval
            of outliers for every iteration. Also, if `normalize` is True,
            every value will be normalized to the [-1, 1] range.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        tdelta_it_mean = _utils.calc_ioe_stats(
            ts=ts,
            funcs=np.mean,
            ts_scaled=ts_scaled,
            step_size=step_size,
            differentiate=differentiate,
        )

        if normalize:
            tdelta_it_mean = 2 * tdelta_it_mean / ts.size - 1

        return tdelta_it_mean

    @classmethod
    def ft_trend_strenght(
        cls,
        ts_residuals: np.ndarray,
        ts_deseasonalized: np.ndarray,
        ddof: int = 1,
    ) -> float:
        """Ratio of standard deviations of time-series and after detrend.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals (random noise) of an one-dimensional time-series.

        ts_deseasonalized: :obj:`np.ndarray`
            One-dimensional deseasonalized time-series values.

        ddof : int, optional (default=1)
            Degrees of freedom for standard deviation.

        Returns
        -------
        float
            Ratio of standard deviation of the original time-series
            and the standard deviation of the detrended version.

        References
        ----------
        .. [1] R. J. Hyndman, E. Wang and N. Laptev, "Large-Scale Unusual Time
            Series Detection," 2015 IEEE International Conference on Data
            Mining Workshop (ICDMW), Atlantic City, NJ, 2015, pp. 1616-1619,
            doi: 10.1109/ICDMW.2015.104.
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        trend = 1.0 - (
            np.var(ts_residuals, ddof=ddof)
            / np.var(ts_deseasonalized, ddof=ddof)
        )

        return min(1.0, max(0.0, trend))

    @classmethod
    def ft_season_strenght(
        cls, ts_residuals: np.ndarray, ts_detrended: np.ndarray, ddof: int = 1
    ) -> float:
        """Ratio of standard deviations of time-series and after deseasoning.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals (random noise) of an one-dimensional time-series.

        ts_deseasonalized: :obj:`np.ndarray`
            One-dimensional deseasonalized time-series values.

        ddof : int, optional (default=1)
            Degrees of freedom for standard deviation.

        Returns
        -------
        float
            Ratio of standard deviation of the original time-series
            and the standard deviation of the deseasonalized version.

        References
        ----------
        .. [1] R. J. Hyndman, E. Wang and N. Laptev, "Large-Scale Unusual Time
            Series Detection," 2015 IEEE International Conference on Data
            Mining Workshop (ICDMW), Atlantic City, NJ, 2015, pp. 1616-1619,
            doi: 10.1109/ICDMW.2015.104.
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """

        seas = 1.0 - (
            np.var(ts_residuals, ddof=ddof) / np.var(ts_detrended, ddof=ddof)
        )

        return min(1.0, max(0.0, seas))

    @classmethod
    def ft_sd_residuals(cls, ts_residuals: np.ndarray, ddof: int = 1) -> float:
        """Compute the standard deviation of the time-series residuals.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals (random noise) of an one-dimensional time-series.

        ddof : int, optional (default=1)
            Degrees of freedom for standard deviation.

        Returns
        -------
        float
            Standard deviation of the time-series residuals.
        """
        return np.std(ts_residuals, ddof=ddof)

    @classmethod
    def ft_sd_diff(
        cls, ts: np.ndarray, num_diff: int = 1, ddof: int = 1
    ) -> float:
        """Standard deviation of the nth-order differenced time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_diff : int, optional (default=1)
            Order of the differentiation.

        ddof : float, optional (default=1)
            Degrees of freedom for standard deviation.

        Returns
        -------
        float
            Standard deviation of the nth-order differenced time-series.
        """
        return np.std(np.diff(ts, n=num_diff), ddof=ddof)

    @classmethod
    def ft_sd_sdiff(
        cls, ts: np.ndarray, ddof: int = 1, ts_period: t.Optional[int] = None
    ) -> float:
        """Seasonal standard dev.  of the first-order differenced time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        ddof : int, optional (default=1)
            Degrees of freedom for standard deviation.

        ts_period : int, optional
            Time-series period. Used to take advantage of precomputations.

        Returns
        -------
        float
            Standard deviation of the first-order difference of the lagged
            time-series by its own period.
        """
        _ts_period = _period.get_ts_period(ts=ts, ts_period=ts_period)
        ts_sdiff = ts[_ts_period:] - ts[:-_ts_period]
        return np.std(ts_sdiff, ddof=ddof)

    @classmethod
    def ft_skewness_residuals(
        cls, ts_residuals: np.ndarray, method: int = 3, adjusted: bool = False
    ) -> float:
        """Compute the skewness of the time-series residuals.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals (random noise) of an one-dimensional time-series.

        method : int, optional (default=3)
            Defines the strategy used for estimate data skewness. This argument
            is used fo compatibility with R package `e1071`. The options must
            be one of the following:

            +--------+-----------------------------------------------+
            |Option  | Formula                                       |
            +--------+-----------------------------------------------+
            |1       | Skew_1 = m_3 / m_2**(3/2)                     |
            |        | (default of ``scipy.stats``)                  |
            +--------+-----------------------------------------------+
            |2       | Skew_2 = Skew_1 * sqrt(n(n-1)) / (n-2)        |
            +--------+-----------------------------------------------+
            |3       | Skew_3 = m_3 / s**3 = Skew_1 ((n-1)/n)**(3/2) |
            +--------+-----------------------------------------------+

            Where `n` is the number of instances in ``ts``, `s` is the standard
            deviation of each attribute in ``ts``, and `m_i` is the ith
            statistical momentum of each attribute in ``ts``.

            Note that if the selected method is unable to be calculated due to
            division by zero, then the first method will be used instead.

        adjusted : bool, optional
            If True, then the calculations are corrected for statistical bias.

        Returns
        -------
        float
            Detrended time-series skewness.

        References
        ----------
        .. [1] Donald Michie, David J. Spiegelhalter, Charles C. Taylor, and
           John Campbell. Machine Learning, Neural and Statistical
           Classification, volume 37. Ellis Horwood Upper Saddle River, 1994.
        """
        ts_skew = _summary.sum_skewness(
            values=ts_residuals, method=method, bias=not adjusted
        )

        return float(ts_skew)

    @classmethod
    def ft_skewness_diff(
        cls,
        ts: np.ndarray,
        num_diff: int = 1,
        method: int = 3,
        adjusted: bool = False,
    ) -> float:
        """Skewness of the nth-order differenced time-series.

        This method calculates the skewness of the nth-order differenced
        time-series (with lag = 1), with `n` being given by `num_diff`.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_diff : int, optional (default=1)
            Order of the differentiation.

        method : int, optional (default=3)
            Defines the strategy used for estimate data skewness. This argument
            is used fo compatibility with R package `e1071`. The options must
            be one of the following:

            +--------+-----------------------------------------------+
            |Option  | Formula                                       |
            +--------+-----------------------------------------------+
            |1       | Skew_1 = m_3 / m_2**(3/2)                     |
            |        | (default of ``scipy.stats``)                  |
            +--------+-----------------------------------------------+
            |2       | Skew_2 = Skew_1 * sqrt(n(n-1)) / (n-2)        |
            +--------+-----------------------------------------------+
            |3       | Skew_3 = m_3 / s**3 = Skew_1 ((n-1)/n)**(3/2) |
            +--------+-----------------------------------------------+

            Where `n` is the number of instances in ``ts``, `s` is the standard
            deviation of each attribute in ``ts``, and `m_i` is the ith
            statistical momentum of each attribute in ``ts``.

            Note that if the selected method is unable to be calculated due to
            division by zero, then the first method will be used instead.

        adjusted : bool, optional
            If True, then the calculations are corrected for statistical bias.

        Returns
        -------
        float
            Skewness of the nth-order differenced time-series

        References
        ----------
        .. [1] Donald Michie, David J. Spiegelhalter, Charles C. Taylor, and
           John Campbell. Machine Learning, Neural and Statistical
           Classification, volume 37. Ellis Horwood Upper Saddle River, 1994.
        """
        ts_diff = np.diff(ts, n=num_diff)
        ts_skew = _summary.sum_skewness(
            values=ts_diff, method=method, bias=not adjusted
        )

        return float(ts_skew)

    @classmethod
    def ft_skewness_sdiff(
        cls,
        ts: np.ndarray,
        method: int = 3,
        adjusted: bool = False,
        ts_period: t.Optional[int] = None,
    ) -> float:
        """Seasonal skewness of the first-order differenced time-series.

        This method calculates the skewness of the first-order differenced
        time-series, lagged with its period.

        If the time-series is not seasonal, then its period is assumed to be 1.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        method : int, optional (default=3)
            Defines the strategy used for estimate data skewness. This argument
            is used fo compatibility with R package `e1071`. The options must
            be one of the following:

            +--------+-----------------------------------------------+
            |Option  | Formula                                       |
            +--------+-----------------------------------------------+
            |1       | Skew_1 = m_3 / m_2**(3/2)                     |
            |        | (default of ``scipy.stats``)                  |
            +--------+-----------------------------------------------+
            |2       | Skew_2 = Skew_1 * sqrt(n(n-1)) / (n-2)        |
            +--------+-----------------------------------------------+
            |3       | Skew_3 = m_3 / s**3 = Skew_1 ((n-1)/n)**(3/2) |
            +--------+-----------------------------------------------+

            Where `n` is the number of instances in ``ts``, `s` is the standard
            deviation of each attribute in ``ts``, and `m_i` is the ith
            statistical momentum of each attribute in ``ts``.

            Note that if the selected method is unable to be calculated due to
            division by zero, then the first method will be used instead.

        adjusted : bool, optional
            If True, then the calculations are corrected for statistical bias.

        ts_period : int, optional
            Time-series period. Used to take advantage of precomputations.

        Returns
        -------
        float
            Skewness of the first-order difference of the lagged time-series
            by its own period.
        """
        _ts_period = _period.get_ts_period(ts=ts, ts_period=ts_period)
        ts_sdiff = ts[_ts_period:] - ts[:-_ts_period]
        ts_skew = _summary.sum_skewness(
            values=ts_sdiff, method=method, bias=not adjusted
        )

        return float(ts_skew)

    @classmethod
    def ft_kurtosis_residuals(
        cls, ts_residuals: np.ndarray, method: int = 3, adjusted: bool = False
    ) -> float:
        """Compute the kurtosis of the time-series residuals.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals (random noise) of an one-dimensional time-series.

        method : int, optional (default=3)
            Defines the strategy used for estimate data kurtosis. Used for
            total compatibility with R package ``e1071``. This option must be
            one of the following:

            +--------+-----------------------------------------------+
            |Method  | Formula                                       |
            +--------+-----------------------------------------------+
            |1       | Kurt_1 = (m_4 / m_2**2 - 3)                   |
            |        | (default of `scipy.stats` package)            |
            +--------+-----------------------------------------------+
            |2       | Kurt_2 = (((n+1) * Kurt_1 + 6) * (n-1) / f_2),|
            |        | f_2 = ((n-2)*(n-3))                           |
            +--------+-----------------------------------------------+
            |3       | Kurt_3 = (m_4 / s**4 - 3)                     |
            |        |        = ((Kurt_1+3) * (1 - 1/n)**2 - 3)      |
            +--------+-----------------------------------------------+

            Where `n` is the number of instances in ``ts``, `s` is the standard
            deviation of each attribute in ``ts``, and `m_i` is the ith
            statistical momentum of each attribute in ``ts``.

            Note that if the selected method is unable to be calculated due
            to division by zero, then the first method is used instead.

        adjusted : bool, optional
            If True, then the calculations are corrected for statistical bias.

        Returns
        -------
        float
            Detrended time-series kurtosis.

        References
        ----------
        .. [1] Donald Michie, David J. Spiegelhalter, Charles C. Taylor, and
           John Campbell. Machine Learning, Neural and Statistical
           Classification, volume 37. Ellis Horwood Upper Saddle River, 1994.
        """
        ts_kurt = _summary.sum_kurtosis(
            values=ts_residuals, method=method, bias=not adjusted
        )

        return float(ts_kurt)

    @classmethod
    def ft_kurtosis_diff(
        cls,
        ts: np.ndarray,
        num_diff: int = 1,
        method: int = 3,
        adjusted: bool = False,
    ) -> float:
        """Kurtosis of the nth-order differenced time-series.

        This method calculates the kurtosis of the nth-order differenced
        time-series (with lag = 1), with `n` being given by `num_diff`.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_diff : int, optional (default=1)
            Order of the differentiation.

        method : int, optional (default=3)
            Defines the strategy used for estimate data kurtosis. Used for
            total compatibility with R package ``e1071``. This option must be
            one of the following:

            +--------+-----------------------------------------------+
            |Method  | Formula                                       |
            +--------+-----------------------------------------------+
            |1       | Kurt_1 = (m_4 / m_2**2 - 3)                   |
            |        | (default of `scipy.stats` package)            |
            +--------+-----------------------------------------------+
            |2       | Kurt_2 = (((n+1) * Kurt_1 + 6) * (n-1) / f_2),|
            |        | f_2 = ((n-2)*(n-3))                           |
            +--------+-----------------------------------------------+
            |3       | Kurt_3 = (m_4 / s**4 - 3)                     |
            |        |        = ((Kurt_1+3) * (1 - 1/n)**2 - 3)      |
            +--------+-----------------------------------------------+

            Where `n` is the number of instances in ``ts``, `s` is the standard
            deviation of each attribute in ``ts``, and `m_i` is the ith
            statistical momentum of each attribute in ``ts``.

            Note that if the selected method is unable to be calculated due
            to division by zero, then the first method is used instead.

        adjusted : bool, optional
            If True, then the calculations are corrected for statistical bias.

        Returns
        -------
        float
            Kurtosis of the nth-order differenced time-series
        """
        ts_diff = np.diff(ts, n=num_diff)
        ts_kurt = _summary.sum_kurtosis(
            values=ts_diff, method=method, bias=not adjusted
        )

        return float(ts_kurt)

    @classmethod
    def ft_kurtosis_sdiff(
        cls,
        ts: np.ndarray,
        method: int = 3,
        adjusted: bool = False,
        ts_period: t.Optional[int] = None,
    ) -> float:
        """Seasonal kurtosis of the first-order differenced time-series.

        This method calculates the kurtosis of the first-order differenced
        time-series, lagged with its period.

        If the time-series is not seasonal, then its period is assumed to be 1.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        method : int, optional (default=3)
            Defines the strategy used for estimate data kurtosis. Used for
            total compatibility with R package ``e1071``. This option must be
            one of the following:

            +--------+-----------------------------------------------+
            |Method  | Formula                                       |
            +--------+-----------------------------------------------+
            |1       | Kurt_1 = (m_4 / m_2**2 - 3)                   |
            |        | (default of `scipy.stats` package)            |
            +--------+-----------------------------------------------+
            |2       | Kurt_2 = (((n+1) * Kurt_1 + 6) * (n-1) / f_2),|
            |        | f_2 = ((n-2)*(n-3))                           |
            +--------+-----------------------------------------------+
            |3       | Kurt_3 = (m_4 / s**4 - 3)                     |
            |        |        = ((Kurt_1+3) * (1 - 1/n)**2 - 3)      |
            +--------+-----------------------------------------------+

            Where `n` is the number of instances in ``ts``, `s` is the standard
            deviation of each attribute in ``ts``, and `m_i` is the ith
            statistical momentum of each attribute in ``ts``.

            Note that if the selected method is unable to be calculated due
            to division by zero, then the first method is used instead.

        adjusted : bool, optional
            If True, then the calculations are corrected for statistical bias.

        ts_period : int, optional
            Time-series period. Used to take advantage of precomputations.

        Returns
        -------
        float
            Kurtosis of the first-order difference of the lagged time-series
            by its own period.
        """
        _ts_period = _period.get_ts_period(ts=ts, ts_period=ts_period)
        ts_sdiff = ts[_ts_period:] - ts[:-_ts_period]
        ts_kurt = _summary.sum_kurtosis(
            values=ts_sdiff, method=method, bias=not adjusted
        )

        return float(ts_kurt)

    @classmethod
    def ft_exp_max_lyap(
        cls, ts: np.ndarray, embed_dim: int = 10, lag: t.Optional[int] = None
    ) -> float:
        """Estimation of the maximum Lyapunov coefficient.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        embed_dim : int, optional (default=10)
            Time-series embed dimension.

        lag : int, optional
            Lag of the embed.

        Returns
        -------
        float
            Estimation of the maximum Lyapunov coefficient.

        References
        ----------
        .. [1] H. E. Hurst, The problem of long-term storage in reservoirs,
            International Association of Scientific Hydrology. Bulletin, vol.
            1, no. 3, pp. 13–27, 1956.
        .. [2] H. E. Hurst, A suggested statistical model of some time series
            which occur in nature, Nature, vol. 180, p. 494, 1957.
        .. [3] R. Weron, Estimating long-range dependence: finite sample
            properties and confidence intervals, Physica A: Statistical
            Mechanics and its Applications, vol. 312, no. 1, pp. 285–299,
            2002.
        .. [4] "nolds" Python package: https://pypi.org/project/nolds/
        .. [5] Lemke, Christiane & Gabrys, Bogdan. (2010). Meta-learning for
            time series forecasting and forecast combination. Neurocomputing.
            73. 2006-2016. 10.1016/j.neucom.2009.09.020.
        """
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", module="nolds", category=RuntimeWarning
            )

            max_lyap_exp = nolds.lyap_r(data=ts, lag=lag, emb_dim=embed_dim)

        return max_lyap_exp

    @classmethod
    def ft_exp_hurst(cls, ts: np.ndarray) -> float:
        """Estimation of the Hurst exponent.

        Check `nolds.hurst_rs` documentation for a clear explanation about
        the underlying function.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        Returns
        -------
        float
            Estimation of the hurst exponent.

        References
        ----------
        .. [1] H. E. Hurst, The problem of long-term storage in reservoirs,
            International Association of Scientific Hydrology. Bulletin, vol.
            1, no. 3, pp. 13–27, 1956.
        .. [2] H. E. Hurst, A suggested statistical model of some time series
            which occur in nature, Nature, vol. 180, p. 494, 1957.
        .. [3] R. Weron, Estimating long-range dependence: finite sample
            properties and confidence intervals, Physica A: Statistical
            Mechanics and its Applications, vol. 312, no. 1, pp. 285–299,
            2002.
        .. [4] "nolds" Python package: https://pypi.org/project/nolds/
        """
        return nolds.hurst_rs(data=ts)

    @classmethod
    def ft_dfa(
        cls, ts: np.ndarray, pol_order: int = 1, overlap_windows: bool = True
    ) -> float:
        """Calculate the Hurst parameter from Detrended fluctuation analysis.

        Note that the ``Hurst parameter`` is not the same quantity as the
        ``Hurst exponent``. The Hurst parameter `H` is defined as the quantity
        such that the following holds: std(ts, l * n) = l ** H * std(ts, n),
        where `ts` is the time-series, `l` is a constant factor, `n` is some
        window length of `ts`, and std(ts, k) is the standard deviation of
        `ts` within a window of size `k`.

        Check `nolds.dfa` documentation for a clear explanation about the
        underlying function.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        pol_order : int, optional (default=1)
            Order of the detrending polynomial within each window of the
            analysis.

        overlap_windows : bool, optional (default=True)
            If True, overlap the windows used while performing the analysis.

        Returns
        -------
        float
            Hurst parameter.

        References
        ----------
        .. [1] C.-K. Peng, S. V. Buldyrev, S. Havlin, M. Simons,
            H. E. Stanley, and A. L. Goldberger, Mosaic organization of
            DNA nucleotides, Physical Review E, vol. 49, no. 2, 1994.
        .. [2] R. Hardstone, S.-S. Poil, G. Schiavone, R. Jansen,
            V. V. Nikulin, H. D. Mansvelder, and K. Linkenkaer-Hansen,
            Detrended fluctuation analysis: A scale-free view on neuronal
            oscillations, Frontiers in Physiology, vol. 30, 2012.
        .. [3] "nolds" Python package: https://pypi.org/project/nolds/
        """
        hurst_coeff = nolds.dfa(ts, order=pol_order, overlap=overlap_windows)
        return hurst_coeff

    @classmethod
    def ft_corr_dim(cls, ts: np.ndarray, emb_dim: int = 1) -> float:
        """Correlation dimension of the time-series.

        It is used the Grassberger-Procaccia algorithm for the correlation
        dimension estimation.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        emb_dim : int, optional (default=1)
            Embedding dimension to estimate the correlation dimension.

        Returns
        -------
        float
            Estimated correlation dimension.

        References
        ----------
        .. [1] P. Grassberger and I. Procaccia, Characterization of strange
            attractors, Physical review letters, vol. 50, no. 5, p. 346,
            1983.
        .. [2] P. Grassberger and I. Procaccia, Measuring the strangeness of
            strange attractors, Physica D: Nonlinear Phenomena, vol. 9,
            no. 1, pp. 189–208, 1983.
        .. [3] P. Grassberger, Grassberger-Procaccia algorithm,
            Scholarpedia, vol. 2, no. 5, p. 3043.
        .. [4] "nolds" Python package. URL: https://pypi.org/project/nolds/
        """
        try:
            corr_dim = nolds.corr_dim(ts, emb_dim=emb_dim)

        except AssertionError:
            corr_dim = np.nan

        return corr_dim

    @classmethod
    def ft_opt_boxcox_coef(
        cls, ts: np.ndarray, adjust_data: bool = True
    ) -> float:
        """Estimated optimal box-cox transformation coefficient.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        adjust_data : bool, optional (default=True)
            If True, transform the data to y(t) = ts(t) - min(ts) + 1. This is
            required for non-positive data. If False, estimate the coefficient
            with the original data, possibly failing if the time-series have
            non-positive data.

        Returns
        -------
        float
            Estimated optimal box-cox transformation coefficient.

        References
        ----------
        .. [1] Box, G. E. P. and Cox, D. R. (1964). An analysis of
            transformations, Journal of the Royal Statistical Society, Series
            B, 26, 211-252.
        .. [2] Yanfei Kang, Rob J. Hyndman, Kate Smith-Miles, Visualising
            forecasting algorithm performance using time series instance
            spaces, International Journal of Forecasting, Volume 33, Issue 2,
            2017, Pages 345-358, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2016.09.004.
        """
        if adjust_data:
            ts = ts - ts.min() + 1

        return scipy.stats.boxcox_normmax(ts, method="mle")

    @classmethod
    def ft_t_mean(cls, ts: np.ndarray, pcut: float = 0.02) -> float:
        """Trimmed mean of the time-series values.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        pcut : float, optional (default=0.02)
            Proportion of outlier cut. Must be in [0, 0.5) range.

        Returns
        -------
        float
            Trimmed mean of time-series.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        return scipy.stats.trim_mean(ts, proportiontocut=pcut)

    @classmethod
    def ft_spikiness(
        cls, ts_residuals: np.ndarray, ddof: int = 1
    ) -> np.ndarray:
        """Spikiness of the time-series residuals.

        The spikiness of the time-series residuals is the variance of the
        variance with jackknife resampling (leave-one-out) on the residuals.
        Here, in order to enable other times of summarization, we return all
        the `jackknifed` variances.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        ddof : int, optional (default=1)
            Degrees of freedom to calculate the variances.

        Returns
        -------
        :obj:`np.ndarray`
            Spikiness of the time-series residuals.

        References
        ----------
        .. [1] R. J. Hyndman, E. Wang and N. Laptev, "Large-Scale Unusual Time
            Series Detection," 2015 IEEE International Conference on Data
            Mining Workshop (ICDMW), Atlantic City, NJ, 2015, pp. 1616-1619,
            doi: 10.1109/ICDMW.2015.104.
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        vars_ = np.array(
            [
                np.var(np.delete(ts_residuals, i), ddof=ddof)
                for i in np.arange(ts_residuals.size)
            ]
        )

        # Note: on the original reference paper, the spikiness is calculated
        # as the variance of the 'vars_'. However, to enable summarization,
        # here we return the full array.
        return vars_



"""Simple model algorithms for time-series forecasting."""
import typing as t
import abc

import numpy as np
import scipy.optimize


class BaseModel(metaclass=abc.ABCMeta):
    """Base model for the custom models of this module."""

    @abc.abstractmethod
    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs) -> "BaseModel":
        """Generic fit method."""

    @abc.abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        """Generic predict method."""


class TSNaive(BaseModel):
    """Naive model for time-series forecasting.

    In the Naive model, all forecasted values are equal to the last known
    observation.
    """

    def __init__(self):
        """Init a Naive model."""
        self.last_obs = -1.0
        self.last_timestamp = -1.0
        self._fitted = False

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs) -> "TSNaive":
        """Fit a Naive model.

        It stores the value of the last observation of ``y``, and its
        timestamp.
        """
        if X.size != y.size:
            raise ValueError("'X' and 'y' size must match.")

        self.last_obs = y[-1]
        self.last_timestamp = X[-1]

        self._fitted = True

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Forecast timestamps ``X``."""
        if not self._fitted:
            raise ValueError("Model is not fitted.")

        if np.any(X < self.last_timestamp):
            raise ValueError(
                "Prediction timestamps 'X' must be all larger "
                "or equal than the last fitted observation "
                "timestamp ({}).".format(self.last_timestamp)
            )

        return np.full(fill_value=self.last_obs, shape=X.shape)


class TSNaiveDrift(BaseModel):
    """Naive model with drift for time-series forecasting.

    In the drift model, the forecasts are equal to the last observation of
    a given time-series plus an additional value proportional to the
    forecasted timestamp. The attributed to the timestamp is estimated from the
    first and last observation of the given time-series.
    """

    def __init__(self):
        """Init a Naive model with drift."""
        self.slope = -1.0
        self.last_obs = -1.0
        self.last_obs_ind = -1
        self._fitted = False

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs) -> "TSNaiveDrift":
        """Fit a Naive model with drift.

        This model calculates the slope of the line crossing the first and last
        observation of ``y``, and stores it alongside the last observation
        value of ``y`` and its timestamp. This is equivalent of calculating the
        mean of the slopes between each pair of adjacent observation, since it
        is a telescoping sum, and use it as the model drift coefficient.
        """
        if X.size != y.size:
            raise ValueError("'X' and 'y' size must match.")

        self.last_obs = y[-1]
        self.last_obs_ind = X[-1]

        # Note: if y.size == 1, this model degenerates to a standard Naive
        # model
        if y.size > 1:
            self.slope = (y[-1] - y[0]) / (X[-1] - X[0])

        else:
            self.slope = 0.0

        self._fitted = True

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict new observations from the timestamps 'X'."""
        if not self._fitted:
            raise ValueError("Model is not fitted.")

        # Note: putting the given timestamps into the perspective of the last
        # value of the fitted time-series, since it is the only reference
        # value of a naive model.
        diff_timestamps = X - self.last_obs_ind

        if np.any(diff_timestamps < 0):
            raise ValueError(
                "Timestamps must be higher than the last fitted "
                "timestamp ({}).".format(self.last_obs_ind)
            )

        return self.last_obs + diff_timestamps * self.slope


class TSNaiveSeasonal(BaseModel):
    """Seasonal Naive model for time-series forecasting.

    This model is similar to the Naive model, but instead of using only the
    very last observation from the fitted time-series, it is used the whole
    past period. Then, each prediction is equal to the value in the
    corresponding timestamp of the previous period.
    """

    def __init__(self, ts_period: int, copy: bool = False):
        """Init a Seasonal Naive Model."""
        self.y = np.empty(0)
        self.ts_period = ts_period
        self.timestamp_interval = -1
        self.last_timestamp = -1
        self._fitted = False
        self.copy = copy

        if self.ts_period is None:
            raise ValueError("'ts_period' must be given.")

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs) -> "TSNaiveSeasonal":
        """Fit a Seasonal Naive model."""
        if X.size != y.size:
            raise ValueError("'X' and 'y' size must match.")

        if self.copy or not isinstance(y, np.ndarray):
            self.y = np.copy(y)

        else:
            self.y = y

        if X.size < self.ts_period:
            raise ValueError(
                "Fitted time-series can't be smaller than its " "period."
            )

        self.timestamp_interval = X[1] - X[0]
        self.last_timestamp = X[-1]

        self._fitted = True

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Make predictions using the Seasonal Naive model."""
        if not self._fitted:
            raise ValueError("Model is not fitted.")

        X = (X / self.timestamp_interval).astype(int)
        shift = (X - self.y.size) // self.ts_period
        inds = X - self.ts_period * (1 + shift)

        if np.any(inds < 0):
            raise ValueError(
                "Timestamps to predict can't be smaller than "
                "the last fitted timestamp."
            )

        return self.y[inds]


class _TSLocalStat(BaseModel):
    """Local statistical forecasting model for time-series.

    This model calculates a statistic from the most recent time-series
    observations, tipically the mean or median, and use the obtained
    value as the forecasted value for future timestamps.
    """

    def __init__(
        self, stat_func: t.Callable[[np.ndarray], float], train_prop: float
    ):
        """Init a Local statistical forecasting model."""
        if not 0 < train_prop <= 1:
            raise ValueError(
                "'train_prop' must be in (0, 1] (got {})."
                "".format(train_prop)
            )

        self.train_prop = train_prop
        self._stat_func = stat_func
        self.loc_mean_fit = -1.0
        self.last_timestamp = -1
        self._fitted = False

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs) -> "_TSLocalStat":
        """Fit a local statistical forecasting model."""
        self.last_timestamp = X[-1]
        last_ind = int(np.ceil(y.size * self.train_prop))
        self.loc_mean_fit = self._stat_func(y[-last_ind:])

        if not np.isscalar(self.loc_mean_fit):
            raise ValueError(
                "Local statistical model demands a function "
                "that return a single scalar value."
            )

        self._fitted = True

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict with the local statistical model."""
        if not self._fitted:
            raise ValueError("Model is not fitted.")

        if np.any(X < self.last_timestamp):
            raise ValueError(
                "Timestamps to predict can't be smaller than "
                "the last fitted timestamp."
            )

        return np.full(fill_value=self.loc_mean_fit, shape=X.shape)


class TSLocalMean(_TSLocalStat):
    """Local mean forecasting model.

    This model calculates the mean from the most recent time-series
    observations, and use the obtained value as the forecasted value
    for future timestamps.
    """

    def __init__(self, train_prop: float = 0.25):
        super().__init__(stat_func=np.mean, train_prop=train_prop)


class TSLocalMedian(_TSLocalStat):
    """Local median forecasting model.

    This model calculates the median from the most recent time-series
    observations, and use the obtained value as the forecasted value
    for future timestamps.
    """

    def __init__(self, train_prop: float = 0.25):
        super().__init__(stat_func=np.median, train_prop=train_prop)


class TSSine(BaseModel):
    """Sine forecasting model.

    The sine model is in the form by y(t) = A * sin(w * t + p) + c, where
    `A`, `w`, `p` and `c` are parameters to be optimized from the fitted
    data.
    """

    def __init__(
        self,
        random_state: t.Optional[int] = None,
        opt_initial_guess: bool = True,
    ):
        """Init the sine forecasting model.

        Parameters
        ----------
        random_state : int, optional
            Random seed, to keep the optimization process deterministic.

        opt_initial_guess : bool, optional (default=True)
            If True, make an informed choice of the initial parameters before
            the optimization process.
        """
        # pylint: disable=C0103
        self.A, self.w, self.p, self.c = 4 * [-1.0]

        self._func = lambda t, A, w, p, c: A * np.sin(w * t + p) + c
        self._fit_func = (
            lambda t: self.A * np.sin(self.w * t + self.p) + self.c
        )

        self.random_state = random_state
        self.opt_initial_guess = opt_initial_guess

        self._fitted = False

    def _make_initial_guess(self, X: np.ndarray, y: np.ndarray) -> np.ndarray:
        """Make the initial guess."""
        # pylint: disable=C0103
        if self.opt_initial_guess:
            # Note: based on: https://stackoverflow.com/a/42322656
            freqs = np.fft.fftfreq(y.size, X[1] - X[0])
            Fyy = np.abs(np.fft.rfft(y))[1:]
            w_guess = 2 * np.pi * np.abs(freqs[1 + np.argmax(Fyy)])
            A_guess = np.std(y) * np.sqrt(2)
            c_guess = np.mean(y)

            return np.asarray([A_guess, w_guess, 0.0, c_guess])

        return np.std(y) * np.random.randn(4)

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs) -> "TSSine":
        """Fit the Sine forecasting model."""
        if self.random_state is not None:
            np.random.seed(self.random_state)

        guess = self._make_initial_guess(X=X, y=y)

        try:
            popt = scipy.optimize.curve_fit(
                self._func, X.ravel(), y, p0=guess, check_finite=False
            )[0]

            self.A, self.w, self.p, self.c = popt

            self._fitted = True

        except RuntimeError:
            self._fitted = False

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Make predictions with the given timestamps ``X``."""
        if not self._fitted:
            raise ValueError("Model is not fitted.")

        return self._fit_func(X)


class TSExp(BaseModel):
    """Exponential forecasting model.

    The exponential model is in the form by y(t) = a * exp(b * t) + c, where
    `a`, `b`, and `c` are parameters to be optimized from the fitted data.
    """

    def __init__(self):
        """Init an exponential forecasting model."""
        # pylint: disable=C0103
        self.a, self.b, self.c = 3 * [-1.0]

        self._func = lambda t, a, b, c: a * np.exp(b * t) + c
        self._fit_func = lambda t: self.a * np.exp(self.b * t) + self.c

        self._fitted = False

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs) -> "TSExp":
        """Fit the exponential forecasting model."""
        b_0 = y[-1] / y[-2]
        a_0 = 0.1
        c_0 = 0
        guess = np.asarray([a_0, b_0, c_0], dtype=float)

        try:
            popt = scipy.optimize.curve_fit(
                self._func, X.ravel(), y, p0=guess, check_finite=False
            )[0]

            self.a, self.b, self.c = popt

            self._fitted = True

        except RuntimeError:
            self._fitted = False

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Make predictions with the given timestamps ``X``."""
        if not self._fitted:
            raise ValueError("Model is not fitted.")

        return self._fit_func(X)



"""Module dedicated to information theoretic time-series meta-features."""
import typing as t

import numpy as np
import sklearn.linear_model
import scipy.stats

import tspymfe._detrend as _detrend
import tspymfe._embed as _embed
import tspymfe._utils as _utils

try:
    import tspymfe.autocorr as autocorr

except ImportError:
    pass


class MFETSInfoTheory:
    """Extract time-series meta-features from Information Theory group."""

    @classmethod
    def precompute_ts_scaled(
        cls, ts: np.ndarray, **kwargs
    ) -> t.Dict[str, np.ndarray]:
        """Precompute a standardized time series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``ts_scaled`` (:obj:`np.ndarray`): standardized time-series
                    values (z-score).
        """
        precomp_vals = {}  # type: t.Dict[str, np.ndarray]

        if "ts_scaled" not in kwargs:
            precomp_vals["ts_scaled"] = _utils.standardize_ts(ts=ts)

        return precomp_vals

    @classmethod
    def precompute_detrended_ami(
        cls,
        ts: np.ndarray,
        num_bins: int = 64,
        lags: t.Optional[t.Union[int, t.Sequence[int]]] = None,
        return_dist: bool = False,
        max_nlags: t.Optional[int] = None,
        adjusted: bool = True,
        **kwargs
    ) -> t.Dict[str, np.ndarray]:
        """Precompute detrended time-series Automutual Information function.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_bins : int, optional (default=64)
            Number of histogram bins to estimate both the probability density
            of each lagged component, and the joint probability distribution,
            which are all necessary to the automutual information computation.

        lags : int or sequence of int, optional
            Lags to calculate the automutual information.
            If int, calculate the automutual information from lag 1 to up to
            the given ``lags`` value.
            If sequence of integers, calculate the automutual information for
            each of the given lags.

        return_dist : bool, optional (default=False)
            If True, return the automutual information distance for every lag,
            defined as:
            $$
                DAMI(ts) = 1 - AMI(ts) / H(ts_A, ts_B)
                         = (H(ts_A) + H(ts_B)) / H(ts_A, ts_B)
            $$

        max_nlags : int, optional
            If ``lags`` is None, then a single lag will be estimated from the
            first negative value of the detrended time-series autocorrelation
            function up to `max_nlags`, if any. Otherwise, lag 1 will be used.
            Used only if ``lags`` is None.

        adjusted : bool, optional (default=True)
            If True, correct the autocorrelation function for statistical
            bias. Used only if ``lags`` is None.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``detrended_ami`` (:obj:`np.ndarray`): the automutual
                    information function of the detrended time-series.

            The following item is necessary and, therefore, also precomputed
            if necessary:
                * ``detrended_acfs`` (:obj:`np.ndarray`): the autocorrelation
                    function from the detrended time-series.

        References
        ----------
        .. [1] Fraser AM, Swinney HL. Independent coordinates for strange
            attractors from mutual information. Phys Rev A Gen Phys.
            1986;33(2):1134‐1140. doi:10.1103/physreva.33.1134
        .. [2] Thomas M. Cover and Joy A. Thomas. 1991. Elements of information
            theory. Wiley-Interscience, USA.
        """
        precomp_vals = {}  # type: t.Dict[str, np.ndarray]

        detrended_acfs = kwargs.get("detrended_acfs")

        if lags is None and detrended_acfs is None:
            precomp_vals.update(
                autocorr.MFETSAutocorr.precompute_detrended_acf(
                    ts=ts, nlags=max_nlags, adjusted=adjusted
                )
            )

            detrended_acfs = precomp_vals["detrended_acfs"]

        if "detrended_ami" not in kwargs:
            precomp_vals["detrended_ami"] = cls.ft_ami_detrended(
                ts=ts,
                lags=lags,
                num_bins=num_bins,
                return_dist=return_dist,
                detrended_acfs=detrended_acfs,
            )

        return precomp_vals

    @classmethod
    def _calc_ami(
        cls,
        ts: np.ndarray,
        lag: int,
        num_bins: int = 64,
        return_dist: bool = False,
    ) -> float:
        """Calculate the Automutual Information of a time-series for a lag.

        The automutual information AMI is defined as:
        $$
            AMI(ts) = H(ts_A) + H(ts_B) - H(ts_A, ts_B)
        $$
        where `ts` is the time-series, $H$ is the Shannon entropy function, and
        $H(A, B)$ is the Shannon entropy of the joint probability of A and B.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        lag : int
            Lag of the automutual information.

        num_bins : int, optional (default=64)
            Number of histogram bins to estimate both the probability density
            of each lagged component, and the joint probability distribution,
            all necessary to the automutual information calculus.

        return_dist : bool, optional (default=False)
            If True, return the automutual information distance defined as
            $$
                DAMI(ts) = 1 - AMI(ts) / H(ts_A, ts_B)
                         = (H(ts_A) + H(ts_B)) / H(ts_A, ts_B)
            $$

        Returns
        -------
        float
            If `return_dist` is False, return the automutual information of
            the time-series with the given lag. If `return_dist` is True,
            return the distance metric version of the automutual information.

        References
        ----------
        .. [1] Fraser AM, Swinney HL. Independent coordinates for strange
            attractors from mutual information. Phys Rev A Gen Phys.
            1986;33(2):1134‐1140. doi:10.1103/physreva.33.1134
        .. [2] Thomas M. Cover and Joy A. Thomas. 1991. Elements of information
            theory. Wiley-Interscience, USA.
        """
        ts_x = ts[:-lag]
        ts_y = ts[lag:]

        ts_x_prob = np.histogram(ts_x, bins=num_bins, density=True)[0]
        ts_y_prob = np.histogram(ts_y, bins=num_bins, density=True)[0]
        joint_prob = np.histogram2d(ts_x, ts_y, bins=num_bins, density=True)[0]

        ent_ts_x = scipy.stats.entropy(ts_x_prob, base=2)
        ent_ts_y = scipy.stats.entropy(ts_y_prob, base=2)
        ent_joint = scipy.stats.entropy(joint_prob.ravel(), base=2)

        ami = ent_ts_x + ent_ts_y - ent_joint

        if return_dist:
            # Note: this is the same as defining, right from the start,
            # ami = (ent_ts_x + ent_ts_y) / ent_joint
            # However, here all steps are kept to make the code clearer.
            ami = 1 - ami / ent_joint

        return ami

    @classmethod
    def ft_hist_entropy(
        cls, ts: np.ndarray, num_bins: int = 10, normalize: bool = True
    ) -> float:
        """Shannon's Entropy from a histogram frequencies.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_bins : int, optional (default=10)
            Number of histogram bins.

        normalize : bool, optional (default=True)
            If True, normalize the result in the [0, 1] range.

        Returns
        -------
        float
            Histogram Shannon's entropy.

        References
        ----------
        .. [1] Shannon, C.E. (1948), A Mathematical Theory of Communication.
            Bell System Technical Journal, 27: 379-423.
            doi:10.1002/j.1538-7305.1948.tb01338.x
        """
        freqs = np.histogram(ts, bins=num_bins, density=True)[0]

        entropy = scipy.stats.entropy(freqs, base=2)

        if normalize:
            entropy /= np.log2(freqs.size)

        return entropy

    @classmethod
    def ft_hist_ent_out_diff(
        cls,
        ts: np.ndarray,
        num_bins: int = 10,
        pcut: float = 0.05,
        normalize: bool = True,
    ) -> float:
        """Difference of histogram entropy with and without outliers.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_bins : int, optional (default=10)
            Number of histogram bins.

        pcut : float, optional (default=0.05)
            Proportion of outlier cut, in both extremes. Must be a value
            in (0.0, 0.5) range.

        normalize : bool, optional (default=True)
            If True, normalize the result in the [0, 1] range.

        Returns
        -------
        float
            Difference of histogram Shannon's entropy with and without outlier
            observations.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        if not 0 < pcut < 0.5:
            raise ValueError(
                "'pcut' must be in (0.0, 0.5) (got " "{}).".format(pcut)
            )

        cut_low, cut_high = np.quantile(ts, (pcut, 1 - pcut))
        ts_inliners = ts[np.logical_and(cut_low <= ts, ts <= cut_high)]

        ent_ts = cls.ft_hist_entropy(
            ts=ts, num_bins=num_bins, normalize=normalize
        )
        ent_ts_inliners = cls.ft_hist_entropy(
            ts=ts_inliners, num_bins=num_bins, normalize=normalize
        )

        entropy_diff = ent_ts - ent_ts_inliners

        return entropy_diff

    @classmethod
    def ft_ami(
        cls,
        ts: np.ndarray,
        num_bins: int = 64,
        lags: t.Optional[t.Sequence[int]] = None,
        return_dist: bool = False,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Automutual information of the time-series.

        The automutual information AMI is defined as:
        $$
            AMI(ts) = H(ts_A) + H(ts_B) - H(ts_A, ts_B)
        $$
        where `ts` is the time-series, $H$ is the Shannon entropy function, and
        $H(A, B)$ is the Shannon entropy of the joint probability of A and B.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_bins : int, optional (default=64)
            Number of histogram bins to estimate both the probability density
            of each lagged component, and the joint probability distribution,
            which are all necessary to the automutual information computation.

        lags : sequence of int, optional
            Lags to calculate the automutual information.

        return_dist : bool, optional (default=False)
            If True, return the automutual information distance for every lag,
            defined as:
            $$
                DAMI(ts) = 1 - AMI(ts) / H(ts_A, ts_B)
                         = (H(ts_A) + H(ts_B)) / H(ts_A, ts_B)
            $$

        max_nlags : int, optional
            If ``lags`` is None, then a single lag will be estimated from the
            first negative value of the detrended time-series autocorrelation
            function up to `max_nlags`, if any. Otherwise, lag 1 will be used.
            Used only if ``detrended_acfs`` is None.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is None.
            If this argument is not given and the previous condiditon is meet,
            the autocorrelation function will be calculated inside this method
            up to ``max_nlags``.

        Returns
        -------
        :obj:`np.ndarray`
            If `return_dist` is False, return the automutual information of
            the time-series for all given lags. If `return_dist` is True,
            return the distance metric version of the automutual information
            for all given lags.

        References
        ----------
        .. [1] Fraser AM, Swinney HL. Independent coordinates for strange
            attractors from mutual information. Phys Rev A Gen Phys.
            1986;33(2):1134‐1140. doi:10.1103/physreva.33.1134
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [4] Thomas M. Cover and Joy A. Thomas. 1991. Elements of information
            theory. Wiley-Interscience, USA.
        """
        # Note: using ts_detrended=ts to avoid detrending.
        non_detrended_ami = cls.ft_ami_detrended(
            ts=ts,
            num_bins=num_bins,
            lags=lags,
            return_dist=return_dist,
            max_nlags=max_nlags,
            detrended_acfs=detrended_acfs,
            ts_detrended=ts,
        )

        return non_detrended_ami

    @classmethod
    def ft_ami_detrended(
        cls,
        ts: np.ndarray,
        num_bins: int = 64,
        lags: t.Optional[t.Union[int, t.Sequence[int]]] = None,
        return_dist: bool = False,
        max_nlags: t.Optional[int] = None,
        ts_detrended: t.Optional[np.ndarray] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Automutual information of the detrended time-series.

        The automutual information AMI is defined as:
        $$
            AMI(ts) = H(ts_A) + H(ts_B) - H(ts_A, ts_B)
        $$
        where `ts` is the time-series, $H$ is the Shannon entropy function, and
        $H(A, B)$ is the Shannon entropy of the joint probability of A and B.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_bins : int, optional (default=64)
            Number of histogram bins to estimate both the probability density
            of each lagged component, and the joint probability distribution,
            which are all necessary to the automutual information computation.

        lags : int or sequence of int, optional
            Lags to calculate the automutual information.
            If int, calculate the automutual information from lag 1 to up to
            the given ``lags`` value.
            If sequence of integers, calculate the automutual information for
            each of the given lags.
            If None, estimate the appropriate lag from the autocorrelation
            function (using the lag corresponding to the first non-positive
            value), and estimate the automutual information from lag 1 up to
            the estimated appropriate lag.

        return_dist : bool, optional (default=False)
            If True, return the automutual information distance for every lag,
            defined as:
            $$
                DAMI(ts) = 1 - AMI(ts) / H(ts_A, ts_B)
                         = (H(ts_A) + H(ts_B)) / H(ts_A, ts_B)
            $$

        max_nlags : int, optional
            If ``lags`` is None, then a single lag will be estimated from the
            first negative value of the detrended time-series autocorrelation
            function up to `max_nlags`, if any. Otherwise, lag 1 will be used.
            Used only if ``detrended_acfs`` is None.

        ts_detrended : :obj:`np.ndarray`, optional
            Detrended time-series. If None, the time-series will be detrended
            using Friedman's Super Smoother.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is None.
            If this argument is not given and the previous condiditon is meet,
            the autocorrelation function will be calculated inside this method
            up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            This method's return value. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            If `return_dist` is False, return the automutual information of
            the time-series for all given lags. If `return_dist` is True,
            return the distance metric version of the automutual information
            for all given lags.

        References
        ----------
        .. [1] Fraser AM, Swinney HL. Independent coordinates for strange
            attractors from mutual information. Phys Rev A Gen Phys.
            1986;33(2):1134‐1140. doi:10.1103/physreva.33.1134
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [4] Thomas M. Cover and Joy A. Thomas. 1991. Elements of information
            theory. Wiley-Interscience, USA.
        """
        if detrended_ami is not None:
            return detrended_ami

        if lags is None:
            lags = _embed.embed_lag(
                ts=ts,
                lag="acf",
                max_nlags=max_nlags,
                detrended_acfs=detrended_acfs,
            )

        if np.isscalar(lags):
            lags = np.arange(1, 1 + int(lags))  # type: ignore

        if ts_detrended is None:
            ts_detrended = _detrend.decompose(ts=ts, ts_period=0)[2]

        _lags: t.Sequence[int] = np.asarray(lags, dtype=int)
        ami = np.zeros(len(_lags), dtype=float)

        for ind, lag in enumerate(_lags):
            ami[ind] = cls._calc_ami(
                ts=ts_detrended,
                lag=lag,
                num_bins=num_bins,
                return_dist=return_dist,
            )

        return ami

    @classmethod
    def ft_ami_first_critpt(
        cls,
        ts: np.ndarray,
        num_bins: int = 64,
        max_nlags: t.Optional[int] = None,
        dist_ami: bool = False,
        detrended_ami: t.Optional[np.ndarray] = None,
    ) -> t.Union[int, float]:
        """First critical point of the automutual information function.

        If `return_dist` is False, then it is search for the first local
        minima in the automutual information function. If `return_dist` is
        True, then search for the first local maxima in the automutual
        information distance function.

        Check `ft_ami` documentation for more information.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_bins : int, optional (default=64)
            Number of histogram bins to estimate both the probability density
            of each lagged component, and the joint probability distribution,
            which are all necessary to the automutual information computation.
            Used only if ``detrended_ami`` is None.

        dist_ami : bool, optional (default=False)
            If True, check for critical points on the automutual information
            distance function. Check `ft_ami` documentation for more
            information.

        max_nlags : int, optional
            Maximum number of lags to estimate the automutual information. Is
            None, then the default value will be min(64, floor(len(ts) / 2)).
            Used only if ``detrended_ami`` is None.

        detrended_ami : :obj:`np.ndarray`, optional
            If `dist_ami` is False, automutual information function from the
            detrended time-series.
            If `dist_ami` is True, automutual information distance function
            from the detrended time-series.
            Used to take advantage of precomputations.

        Returns
        -------
        int or float
            If `dist_ami` False, lag corresponding to the first local minima
            in the automutual information function. If `dist_ami` is True, lag
            corresponding to the first local maxima of the automutual
            information distance function. If no critical point of interest is
            found, this method will return `np.nan`.

        References
        ----------
        .. [1] Fraser AM, Swinney HL. Independent coordinates for strange
            attractors from mutual information. Phys Rev A Gen Phys.
            1986;33(2):1134‐1140. doi:10.1103/physreva.33.1134
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [4] Thomas M. Cover and Joy A. Thomas. 1991. Elements of information
            theory. Wiley-Interscience, USA.
        """
        if detrended_ami is None:
            if max_nlags is None:
                max_nlags = min(64, ts.size // 2)

            detrended_ami = cls.ft_ami_detrended(
                ts=ts, num_bins=num_bins, lags=max_nlags, return_dist=dist_ami
            )

        # Note: if 'return_dist=True', return the first local maximum.
        # If otherwise, return the first local minimum.
        type_ = "max" if dist_ami else "min"

        crit_point = _utils.find_crit_pt(arr=detrended_ami, type_=type_)

        try:
            return np.flatnonzero(crit_point)[0] + 1

        except IndexError:
            return np.nan

    @classmethod
    def ft_ami_curvature(
        cls,
        ts: np.ndarray,
        noise_range: t.Tuple[float, float] = (0, 3),
        noise_inc_num: float = 10,
        lag: t.Optional[t.Union[str, int]] = None,
        random_state: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
    ) -> float:
        """Estimate the Automutual information curvature.

        The Automutual information curvature is estimated using iterative noise
        amplification strategy.

        In the iterative noise amplification strategy, a random white noise
        is sampled from a normal distribution (mean 0 and variance 1). Then,
        this same noise is iteratively amplified from a uniformly spaced
        scales in ``noise_range`` range and added to the time-series. The
        automutual information is calculated from the perturbed time-series
        for each noise amplification.

        The automutual information curvature is the angular coefficient of a
        linear regression of the automutual information onto the noise scales.

        The lag used for every iteration is fixed from the start and, if not
        fixed by the user, it is estimated from the autocorrelation function
        by default.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        noise_range : tuple of float, optional (default=(0, 3))
            A tuple of floats in the form (min_scale, max_scale) for the noise
            amplication range.

        noise_inc_num: float, optional (default=10)
            Number of noise amplifications. The parameter ``noise_range`` will
            be split evenly into ``noise_inc_num`` parts.

        lag : int or str, optional
            Lag to calculate the statistic. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        random_state : int, optional
            Random seed to ensure reproducibility.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        max_nlags : int, optional
            If ``lag`` is None, then a single lag will be estimated from the
            first negative value of the detrended time-series autocorrelation
            function up to `max_nlags`, if any. Otherwise, lag 1 will be used.
            Used only if ``detrended_acfs`` is None.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is None.
            If this argument is not given and the previous condiditon is meet,
            the autocorrelation function will be calculated inside this method
            up to ``max_nlags``.

        Returns
        -------
        float
            Estimated automutual information curvature.

        References
        ----------
        .. [1] Fraser AM, Swinney HL. Independent coordinates for strange
            attractors from mutual information. Phys Rev A Gen Phys.
            1986;33(2):1134‐1140. doi:10.1103/physreva.33.1134
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [4] Thomas M. Cover and Joy A. Thomas. 1991. Elements of information
            theory. Wiley-Interscience, USA.
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        # Note: casting lag to an array since 'ft_ami_detrended' demands
        # a sequence of lags.
        _lag = np.asarray(
            [
                _embed.embed_lag(
                    ts=ts_scaled,
                    lag=lag,
                    max_nlags=max_nlags,
                    detrended_acfs=detrended_acfs,
                )
            ]
        )

        if random_state is not None:
            np.random.seed(random_state)

        # Note: the noise is fixed from the start, and amplified at each
        # iteration.
        gaussian_noise = np.random.randn(ts_scaled.size)
        noise_std = np.linspace(*noise_range, noise_inc_num)

        ami = np.zeros(noise_inc_num, dtype=float)

        for ind, cur_std in enumerate(noise_std):
            ts_corrupted = ts_scaled + cur_std * gaussian_noise

            ami[ind] = cls.ft_ami_detrended(
                ts=ts_corrupted, num_bins=32, lags=_lag, return_dist=False
            )

        model = sklearn.linear_model.LinearRegression().fit(
            X=noise_std.reshape(-1, 1), y=ami
        )

        curvature = model.coef_[0]

        return curvature

    @classmethod
    def ft_approx_entropy(
        cls,
        ts: np.ndarray,
        embed_dim: int = 2,
        embed_lag: int = 1,
        threshold: float = 0.2,
        metric: str = "chebyshev",
        p: t.Union[int, float] = 2,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> float:
        """Approximate entropy of the time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        embed_dim : int, optional (default=2)
            Embedding dimension.

        embed_dim : int, optional (default=1)
            Embedding lag.

        threshold : float, optional (default=0.2)
            Threshold to consider which observations are next to each other
            after embedding.

        metric : str, optional (default="chebyshev")
            Distance metric to calculate the pairwise distance of the
            observations after each embedding.
            Check `scipy.spatial.distance.cdist` documentation for the complete
            list of available distance metrics.

        p : int or float, optional (default=2)
            Power parameter for the minkowski metric. Used only if metric is
            `minkowski`.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        float
            Estimated approximate entropy.

        References
        ----------
        .. [1] Pincus, S.M., Gladstone, I.M. & Ehrenkranz, R.A. A regularity
            statistic for medical data analysis. J Clin Monitor Comput 7,
            335–345 (1991). https://doi.org/10.1007/BF01619355
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """

        def neigh_num(dim: int) -> int:
            """Mean-log-mean of the number of radius neighbors."""
            embed = _embed.embed_ts(ts_scaled, dim=dim, lag=embed_lag)
            dist_mat = scipy.spatial.distance.cdist(
                embed, embed, metric=metric, p=p
            )
            return np.mean(np.log(np.mean(dist_mat < threshold, axis=1)))

        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        approx_entropy = neigh_num(embed_dim) - neigh_num(embed_dim + 1)

        return approx_entropy

    @classmethod
    def ft_sample_entropy(
        cls,
        ts: np.ndarray,
        embed_dim: int = 2,
        embed_lag: int = 1,
        threshold: float = 0.2,
        metric: str = "chebyshev",
        p: t.Union[int, float] = 2,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> float:
        """Sample entropy of the time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        embed_dim : int, optional (default=2)
            Embedding dimension.

        embed_dim : int, optional (default=1)
            Embedding lag.

        threshold : float, optional (default=0.2)
            Threshold to consider which observations are next to each other
            after embedding.

        metric : str, optional (default="chebyshev")
            Distance metric to calculate the pairwise distance of the
            observations after each embedding.
            Check `scipy.spatial.distance.cdist` documentation for the complete
            list of available distance metrics.

        p : int or float, optional (default=2)
            Power parameter for the minkowski metric. Used only if metric is
            `minkowski`.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        float
            Estimated sample entropy.

        References
        ----------
        .. [1] Physiological time-series analysis using approximate entropy and
            sample entropy Joshua S. Richman and J. Randall Moorman, American
            Journal of Physiology-Heart and Circulatory Physiology 2000 278:6,
            H2039-H2049
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """

        def log_neigh_num(dim: int) -> int:
            """Logarithm of the number of nearest neighbors."""
            embed = _embed.embed_ts(ts_scaled, dim=dim, lag=embed_lag)
            dist_mat = scipy.spatial.distance.pdist(embed, metric=metric, p=p)
            return np.log(np.sum(dist_mat < threshold))

        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        sample_entropy = log_neigh_num(embed_dim) - log_neigh_num(
            embed_dim + 1
        )

        return sample_entropy

    @classmethod
    def ft_control_entropy(
        cls,
        ts: np.ndarray,
        embed_dim: int = 2,
        threshold: float = 0.2,
        metric: str = "chebyshev",
        p: t.Union[int, float] = 2,
        embed_lag: int = 1,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> float:
        """Control entropy of the time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        embed_dim : int, optional (default=2)
            Embedding dimension.

        embed_dim : int, optional (default=1)
            Embedding lag.

        threshold : float, optional (default=0.2)
            Threshold to consider which observations are next to each other
            after embedding.

        metric : str, optional (default="chebyshev")
            Distance metric to calculate the pairwise distance of the
            observations after each embedding.
            Check `scipy.spatial.distance.cdist` documentation for the complete
            list of available distance metrics.

        p : int or float, optional (default=2)
            Power parameter for the minkowski metric. Used only if metric is
            `minkowski`.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        float
            Estimated control entropy.

        References
        ----------
        .. [1] Erik M. Bollt and Joseph D. Skufca, Control Entropy: a
            complexity measure for nonstationary signals Mathematical
            Biosciences and Engineering Volume 6, Number 1, January 2009
            doi:10.3934/mbe.2009.6.1 pp. 1–25
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        control_entropy = cls.ft_sample_entropy(
            ts=np.diff(ts),
            embed_dim=embed_dim,
            embed_lag=embed_lag,
            threshold=threshold,
            metric=metric,
            p=p,
            ts_scaled=ts_scaled,
        )

        return control_entropy

    @classmethod
    def ft_surprise(
        cls,
        ts: np.ndarray,
        num_bins: int = 10,
        memory_size: t.Union[float, int] = 0.1,
        num_it: int = 128,
        method: str = "distribution",
        diff_order: int = 1,
        epsilon: float = 1e-8,
        random_state: t.Optional[int] = None,
    ) -> np.ndarray:
        """Surprise factor of a nth-order differenced time-series.

        The surprise measure is an estimation of the negative log-probablity of
        a given random reference observation have its value given a short-term
        memory of `ceil(memory_size * len(ts))` most recent values.

        This method measures the surprise factor associated with each value on
        a nth-order differenced time-series. This analysis may also be
        performed on the original time-series, simply using the 0th-order
        differenced time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_bins : int, optional (default=10)
            Number of histogram bins to time-series discretization. The data is
            discretized using an equiprobable histogram (i.e., all bins have
            the same number of instances).

        memory_size : float or int, optional (default=0.1)
            Recent memory size S.
            If 0 < S < 1, the proportion of the time-series length as the
            memory size.
            If S >= 1, length of the memory.
            Must be a value stricly positive and, if ``method`` value is
            `1-transition`, then it must be larger or equal than 2.

        num_it : int, optional (default=128)
            Number of reference observations sampled.

        method : str, optional (default="distribution")
            Defines the probability function to determine how much surprise
            a given value causes given the past values in the recent memory.
            Must be either `distribution` or `1-transition`.
            1. `distribution`: The ``memory_size`` previous values immediately
                before the reference value are compared to it.
            2. `1-transition`: compare the current reference value to the
                values one time unit after the values equal to the previous
                value of the current reference value within the recent memory.
                This means that we are searching for transitions `B -> A`, from
                the previous value `B` to the current value `A`, in the recent
                memory, and calculating the probability of that event happens.
                In other words, we are calculating the conditional probability
                P(y[t] = A | y[t-1] = B).

        diff_order : int, optional (default=1)
            Order of differentiation of the time-series to carry the analysis.
            If `0`, then this analysis will be performed on the original
            time-series value. Must be a non-negative integer.

        epsilon : float, optional (default=1e-8)
            Tiny threshold value to consider probabilities as zero.

        random_state : int, optional
            Random seed to ensure reproducibility.

        Returns
        -------
        :obj:`np.ndarray`
            Surprise (-1.0 * log-probability) for each random reference
            instance.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        VALID_METHODS = ("distribution", "1-transition")

        if method not in VALID_METHODS:
            raise ValueError(
                "'method' must in {} (got '{}')."
                "".format(VALID_METHODS, method)
            )

        if memory_size <= 0:
            raise ValueError(
                "'memory_size' must be positive (got "
                "{}).".format(memory_size)
            )

        if num_it <= 0:
            raise ValueError(
                "'num_it' must be positive (got {})." "".format(num_it)
            )

        def get_reference_inds(inds_num: int, max_ind: int) -> np.ndarray:
            """Get min(ts.size - memory_size, inds_num) random indices.

            The indices are chosen randomly without replacement.
            """
            if max_ind - memory_size > num_it:
                if random_state is not None:
                    np.random.seed(random_state)

                # Note: we can't use indices before the first 'memory_size'
                # indices as reference, since we need a 'memory' of size
                # 'memory_size' (i.e., we need at least 'memory_size' past
                # indices). Therefore, we always skip they.
                return memory_size + np.random.choice(
                    max_ind - memory_size, size=inds_num, replace=False
                )

            # Note: the requested number of indices is not smaller than
            # the number of available indices. Therefore, just return
            # all available indices.
            return np.arange(memory_size, max_ind)

        if 0 < memory_size < 1:
            memory_size = int(np.ceil(ts.size * memory_size))

        if method == "distribution":

            def prob_func(
                ref_ind: int, ts_bin: np.ndarray, hor_len: int
            ) -> float:
                """Return probability of the referenced value."""
                return np.mean(
                    ts_bin[ref_ind - hor_len : ref_ind] == ts_bin[ref_ind]
                )

        else:
            if memory_size <= 1:
                raise ValueError(
                    "'memory_size' must be >= 2 with " "'1-transition' method."
                )

            def prob_func(
                ref_ind: int, ts_bin: np.ndarray, hor_len: int
            ) -> float:
                """Return probability of the referenced transition."""
                mem_data = ts_bin[ref_ind - hor_len : ref_ind]
                prev_val = mem_data[-1]
                prev_val_inds = np.flatnonzero(mem_data[:-1] == prev_val)
                equal_vals = mem_data[prev_val_inds + 1] == ts_bin[ref_ind]
                return np.mean(equal_vals) if equal_vals.size else 0.0

        ts_diff = np.diff(ts, n=diff_order)

        # Note: discretize time-series using an equiprobable histogram
        # (i.e. all bins have approximately the same number of instances).
        ts_bin = _utils.discretize(
            ts=ts_diff, num_bins=num_bins, strategy="equiprobable"
        )

        probs = np.zeros(num_it, dtype=float)
        ref_inds = get_reference_inds(inds_num=num_it, max_ind=ts_diff.size)

        for ind, ref_ind in enumerate(ref_inds):
            probs[ind] = prob_func(
                ref_ind=ref_ind, ts_bin=ts_bin, hor_len=int(memory_size)
            )

        probs[probs < epsilon] = 1.0
        surprise = -np.log(probs)

        return surprise

    @classmethod
    def ft_lz_complexity(
        cls, ts: np.ndarray, num_bins: int = 10, normalize: bool = True
    ) -> float:
        """Lempel-Ziv complexity of the discretized time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_bins : int, optional (default=10)
            Number of histogram bins to discretize the time-series. It is
            used a histogram with bins of equal width.

        normalize : bool, optional (default=True)
            If True, normalize the final measure in [0, 1] range.

        Returns
        -------
        float
            If `normalize` is False, Lempel-Ziv complexity estimation for the
            discretized time-series. If `normalize` is True, then the measure
            will be normalized accordingly to the number of histogram bins
            used.

        References
        ----------
        .. [1] A. Lempel and J. Ziv, "On the Complexity of Finite Sequences,"
            in IEEE Transactions on Information Theory, vol. 22, no. 1, pp.
            75-81, January 1976, doi: 10.1109/TIT.1976.1055501.
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        ts_bin = tuple(
            _utils.discretize(
                ts=ts, num_bins=num_bins, strategy="equal-width", dtype=int
            )
        )

        ind_start, ind_end = 0, 1
        substrings = set()
        _len = len(ts_bin)

        while ind_end <= _len:
            substring = ts_bin[ind_start:ind_end]

            if substring not in substrings:
                substrings.add(substring)
                ind_start = ind_end

            ind_end += 1

        lz_comp = len(substrings)

        if normalize:
            lz_comp *= np.log(_len) / (_len * np.log(num_bins))

        return lz_comp
